"""Contains the ObjectiveCreator class."""

from __future__ import annotations

import logging
import numbers
import os
import re
import shutil
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass
from typing import (
    Any,
)

import numpy as np
import pandas as pd
import petab.v1 as petab
from petab.v1.C import (
    OBSERVABLE_FORMULA,
    PREEQUILIBRATION_CONDITION_ID,
    SIMULATION_CONDITION_ID,
)
from petab.v1.models import MODEL_TYPE_SBML
from petab.v1.models.sbml_model import SbmlModel
from petab.v1.parameter_mapping import ParMappingDictQuadruple
from petab.v1.simulate import Simulator

from ..C import CENSORED, CONDITION_SEP, LIN, ORDINAL, SEMIQUANTITATIVE
from ..hierarchical.inner_calculator_collector import InnerCalculatorCollector
from ..objective import AmiciObjective, ObjectiveBase, PetabSimulatorObjective
from ..objective.amici import AmiciObjectBuilder
from ..objective.roadrunner import (
    ExpData,
    RoadRunnerCalculator,
    RoadRunnerObjective,
)
from ..predict import AmiciPredictor
from ..result import PredictionResult

try:
    import amici
    import amici.sim.sundials as asd
except ImportError:
    amici = None
try:
    import libsbml
    import roadrunner
except ImportError:
    roadrunner = None
    libsbml = None

logger = logging.getLogger(__name__)


class ObjectiveCreator(ABC):
    """Abstract Creator for creating an objective function."""

    @abstractmethod
    def create_objective(self, **kwargs) -> ObjectiveBase:
        """Create an objective function."""
        pass


class AmiciObjectiveCreator(ObjectiveCreator, AmiciObjectBuilder):
    """ObjectiveCreator for creating an amici objective function."""

    def __init__(
        self,
        petab_problem: petab.Problem,
        hierarchical: bool = False,
        non_quantitative_data_types: Iterable[str] | None = None,
        inner_options: dict[str, Any] | None = None,
        output_folder: str | None = None,
        model_name: str | None = None,
        validate_petab: bool = True,
    ):
        """
        Initialize the creator.

        Parameters
        ----------
        petab_problem:
            The PEtab problem.
        hierarchical:
            Whether to use hierarchical optimization.
        non_quantitative_data_types:
            The non-quantitative data types to consider.
        inner_options:
            Options for the inner optimization.
        output_folder:
            The output folder for the compiled model.
        model_name:
            The name of the model.
        validate_petab:
            Whether to check the PEtab problem for errors.
        """
        self.petab_problem = petab_problem
        self._hierarchical = hierarchical
        self._non_quantitative_data_types = non_quantitative_data_types
        self.inner_options = inner_options
        self.output_folder = output_folder
        self.model_name = model_name
        self.validate_petab = validate_petab

    def create_model(
        self,
        force_compile: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> asd.Model:
        """
        Import amici model.

        Parameters
        ----------
        force_compile:
            If False, the model is compiled only if the output folder does not
            exist yet. If True, the output folder is deleted and the model
            (re-)compiled in either case.

            .. warning::
                If `force_compile`, then an existing folder of that name will
                be deleted.
        verbose:
            Passed to AMICI's model compilation. If True, the compilation
            progress is printed.
        kwargs:
            Extra arguments passed to amici.SbmlImporter.sbml2amici
        """
        # courtesy check whether target is folder
        if os.path.exists(self.output_folder) and not os.path.isdir(
            self.output_folder
        ):
            raise AssertionError(
                f"Refusing to remove {self.output_folder} for model "
                f"compilation: Not a folder."
            )

        # compile
        if self._must_compile(force_compile):
            logger.info(
                f"Compiling amici model to folder {self.output_folder}."
            )
            if self.petab_problem.model.type_id == MODEL_TYPE_SBML:
                self.compile_model(
                    validate=self.validate_petab,
                    verbose=verbose,
                    **kwargs,
                )
            else:
                self.compile_model(verbose=verbose, **kwargs)
        else:
            logger.debug(
                f"Using existing amici model in folder {self.output_folder}."
            )

        return self._create_model()

    def _create_model(self) -> asd.Model:
        """Load model module and return the model, no checks/compilation."""
        import amici
        from amici.importers.petab.v1._import_helpers import check_model

        # load moduĺe
        module = amici.import_model_module(
            module_name=self.model_name, module_path=self.output_folder
        )
        model = module.get_model()
        check_model(
            amici_model=model,
            petab_problem=self.petab_problem,
        )

        return model

    def _must_compile(self, force_compile: bool):
        """Check whether the model needs to be compiled first."""
        # asked by user
        if force_compile:
            return True

        # folder does not exist
        if not os.path.exists(self.output_folder) or not os.listdir(
            self.output_folder
        ):
            return True

        # try to import (in particular checks version)
        try:
            # importing will already raise an exception if version wrong
            amici.import_model_module(self.model_name, self.output_folder)
        except ModuleNotFoundError:
            return True
        except amici.AmiciVersionError as e:
            logger.info(
                f"amici model will be re-imported due to version mismatch: {e}"
            )
            return True

        # no need to (re-)compile
        return False

    def compile_model(self, **kwargs):
        """
        Compile the model.

        If the output folder exists already, it is first deleted.

        Parameters
        ----------
        kwargs:
            Extra arguments passed to :meth:`amici.sbml_import.SbmlImporter.sbml2amici`
            or :func:`amici.pysb_import.pysb2amici`.
        """
        from amici.importers.petab.v1 import import_petab_problem

        # delete output directory
        if os.path.exists(self.output_folder):
            shutil.rmtree(self.output_folder)

        import_petab_problem(
            petab_problem=self.petab_problem,
            model_name=self.model_name,
            output_dir=self.output_folder,
            **kwargs,
        )

    def create_solver(
        self,
        model: asd.Model = None,
        verbose: bool = True,
    ) -> asd.Solver:
        """Return model solver."""
        # create model
        if model is None:
            model = self.create_model(verbose=verbose)

        solver = model.create_solver()
        return solver

    def create_edatas(
        self,
        model: asd.Model = None,
        simulation_conditions=None,
        verbose: bool = True,
    ) -> list[asd.ExpData]:
        """Create list of :class:`amici.amici.ExpData` objects."""
        from amici.sim.sundials.petab.v1 import create_edatas

        # create model
        if model is None:
            model = self.create_model(verbose=verbose)

        return create_edatas(
            amici_model=model,
            petab_problem=self.petab_problem,
            simulation_conditions=simulation_conditions,
        )

    def create_objective(
        self,
        model: asd.Model = None,
        solver: asd.Solver = None,
        edatas: Sequence[asd.ExpData] = None,
        force_compile: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> AmiciObjective:
        """Create a :class:`pypesto.objective.AmiciObjective`.

        Parameters
        ----------
        model:
            The AMICI model.
        solver:
            The AMICI solver.
        edatas:
            The experimental data in AMICI format.
        force_compile:
            Whether to force-compile the model if not passed.
        verbose:
            Passed to AMICI's model compilation. If True, the compilation
            progress is printed.
        **kwargs:
            Additional arguments passed on to the objective. In case of ordinal
            or semiquantitative measurements, ``inner_options`` can optionally
            be passed here. If none are given, ``inner_options`` given to the
            importer constructor (or inner defaults) will be chosen.

        Returns
        -------
        A :class:`pypesto.objective.AmiciObjective` for the model and the data.
        """
        from amici.sim.sundials.petab.v1 import fill_in_parameters
        from amici.sim.sundials.petab.v1._parameter_mapping import (
            create_parameter_mapping,
        )

        simulation_conditions = petab.get_simulation_conditions(
            self.petab_problem.measurement_df
        )
        if model is None:
            model = self.create_model(
                force_compile=force_compile, verbose=verbose
            )
        if solver is None:
            solver = self.create_solver(model)
        # create conditions and edatas from measurement data
        if edatas is None:
            edatas = self.create_edatas(
                model=model, simulation_conditions=simulation_conditions
            )
        parameter_mapping = create_parameter_mapping(
            petab_problem=self.petab_problem,
            simulation_conditions=simulation_conditions,
            scaled_parameters=True,
            amici_model=model,
            fill_fixed_parameters=False,
        )
        par_ids = self.petab_problem.x_ids

        # fill in dummy parameters (this is needed since some objective
        #  initialization e.g. checks for preeq parameters)
        problem_parameters = dict(
            zip(
                self.petab_problem.x_ids,
                self.petab_problem.x_nominal_scaled,
                strict=True,
            )
        )
        fill_in_parameters(
            edatas=edatas,
            problem_parameters=problem_parameters,
            scaled_parameters=True,
            parameter_mapping=parameter_mapping,
            amici_model=model,
        )

        calculator = None
        amici_reporting = None

        if (
            self._non_quantitative_data_types is not None
            and self._hierarchical
        ):
            inner_options = kwargs.pop("inner_options", None)
            inner_options = (
                inner_options
                if inner_options is not None
                else self.inner_options
            )
            calculator = InnerCalculatorCollector(
                self._non_quantitative_data_types,
                self.petab_problem,
                model,
                edatas,
                inner_options,
            )
            amici_reporting = asd.RDataReporting.full

            # FIXME: currently not supported with hierarchical
            if "guess_steadystate" in kwargs and kwargs["guess_steadystate"]:
                warnings.warn(
                    "`guess_steadystate` not supported with hierarchical "
                    "optimization. Disabling `guess_steadystate`.",
                    stacklevel=1,
                )
            kwargs["guess_steadystate"] = False
            inner_parameter_ids = calculator.get_inner_par_ids()
            par_ids = [x for x in par_ids if x not in inner_parameter_ids]

        max_sensi_order = kwargs.get("max_sensi_order", None)

        if (
            self._non_quantitative_data_types is not None
            and any(
                data_type in self._non_quantitative_data_types
                for data_type in [ORDINAL, CENSORED, SEMIQUANTITATIVE]
            )
            and max_sensi_order is not None
            and max_sensi_order > 1
        ):
            raise ValueError(
                "Ordinal, censored and semiquantitative data cannot be "
                "used with second order sensitivities. Use a up to first order "
                "method or disable ordinal, censored and semiquantitative "
            )

        # create objective
        obj = AmiciObjective(
            amici_model=model,
            amici_solver=solver,
            edatas=edatas,
            x_ids=par_ids,
            x_names=par_ids,
            parameter_mapping=parameter_mapping,
            amici_object_builder=self,
            calculator=calculator,
            amici_reporting=amici_reporting,
            **kwargs,
        )

        return obj

    def create_predictor(
        self,
        objective: AmiciObjective = None,
        amici_output_fields: Sequence[str] = None,
        post_processor: Callable | None = None,
        post_processor_sensi: Callable | None = None,
        post_processor_time: Callable | None = None,
        max_chunk_size: int | None = None,
        output_ids: Sequence[str] = None,
        condition_ids: Sequence[str] = None,
    ) -> AmiciPredictor:
        """Create a :class:`pypesto.predict.AmiciPredictor`.

        The `AmiciPredictor` facilitates generation of predictions from
        parameter vectors.

        Parameters
        ----------
        objective:
            An objective object, which will be used to get model simulations
        amici_output_fields:
            keys that exist in the return data object from AMICI, which should
            be available for the post-processors
        post_processor:
            A callable function which applies postprocessing to the simulation
            results. Default are the observables of the AMICI model.
            This method takes a list of ndarrays (as returned in the field
            ['y'] of amici ReturnData objects) as input.
        post_processor_sensi:
            A callable function which applies postprocessing to the
            sensitivities of the simulation results. Default are the
            observable sensitivities of the AMICI model.
            This method takes two lists of ndarrays (as returned in the
            fields ['y'] and ['sy'] of amici ReturnData objects) as input.
        post_processor_time:
            A callable function which applies postprocessing to the timepoints
            of the simulations. Default are the timepoints of the amici model.
            This method takes a list of ndarrays (as returned in the field
            ['t'] of amici ReturnData objects) as input.
        max_chunk_size:
            In some cases, we don't want to compute all predictions at once
            when calling the prediction function, as this might not fit into
            the memory for large datasets and models.
            Here, the user can specify a maximum number of conditions, which
            should be simulated at a time.
            Default is 0 meaning that all conditions will be simulated.
            Other values are only applicable, if an output file is specified.
        output_ids:
            IDs of outputs, if post-processing is used
        condition_ids:
            IDs of conditions, if post-processing is used

        Returns
        -------
        A :class:`pypesto.predict.AmiciPredictor` for the model, using
        the outputs of the AMICI model and the timepoints from the PEtab data.
        """
        # if the user didn't pass an objective function, we create it first
        if objective is None:
            objective = self.create_objective()

        # create a identifiers of preequilibration and simulation condition ids
        # which can then be stored in the prediction result
        edata_conditions = objective.amici_object_builder.petab_problem.get_simulation_conditions_from_measurement_df()
        if PREEQUILIBRATION_CONDITION_ID not in list(edata_conditions.columns):
            preeq_dummy = [""] * edata_conditions.shape[0]
            edata_conditions[PREEQUILIBRATION_CONDITION_ID] = preeq_dummy
        edata_conditions.drop_duplicates(inplace=True)

        if condition_ids is None:
            condition_ids = [
                edata_conditions.loc[id, PREEQUILIBRATION_CONDITION_ID]
                + CONDITION_SEP
                + edata_conditions.loc[id, SIMULATION_CONDITION_ID]
                for id in edata_conditions.index
            ]

        # wrap around AmiciPredictor
        predictor = AmiciPredictor(
            amici_objective=objective,
            amici_output_fields=amici_output_fields,
            post_processor=post_processor,
            post_processor_sensi=post_processor_sensi,
            post_processor_time=post_processor_time,
            max_chunk_size=max_chunk_size,
            output_ids=output_ids,
            condition_ids=condition_ids,
        )

        return predictor

    def rdatas_to_measurement_df(
        self,
        rdatas: Sequence[asd.ReturnData],
        model: asd.Model = None,
        verbose: bool = True,
    ) -> pd.DataFrame:
        """
        Create a measurement dataframe in the petab format.

        Parameters
        ----------
        rdatas:
            A list of rdatas as produced by
            ``pypesto.AmiciObjective.__call__(x, return_dict=True)['rdatas']``.
        model:
            The amici model.
        verbose:
            Passed to AMICI's model compilation. If True, the compilation
            progress is printed.

        Returns
        -------
        A dataframe built from the rdatas in the format as in
        ``self.petab_problem.measurement_df``.
        """
        from amici.sim.sundials.petab.v1 import (
            rdatas_to_measurement_df,
        )

        # create model
        if model is None:
            model = self.create_model(verbose=verbose)

        return rdatas_to_measurement_df(
            rdatas, model, self.petab_problem.measurement_df
        )

    def rdatas_to_simulation_df(
        self,
        rdatas: Sequence[asd.ReturnData],
        model: asd.Model = None,
    ) -> pd.DataFrame:
        """
        See :meth:`rdatas_to_measurement_df`.

        Except a petab simulation dataframe is created, i.e. the measurement
        column label is adjusted.
        """
        return self.rdatas_to_measurement_df(rdatas, model).rename(
            columns={petab.MEASUREMENT: petab.SIMULATION}
        )

    def prediction_to_petab_measurement_df(
        self,
        prediction: PredictionResult,
        predictor: AmiciPredictor = None,
    ) -> pd.DataFrame:
        """
        Cast prediction into a dataframe.

        If a PEtab problem is simulated without post-processing, then the
        result can be cast into a PEtab measurement or simulation dataframe

        Parameters
        ----------
        prediction:
            A prediction result as produced by an :class:`pypesto.predict.AmiciPredictor`.
        predictor:
            The :class:`pypesto.predict.AmiciPredictor` instance.

        Returns
        -------
        A dataframe built from the rdatas in the format as in
        ``self.petab_problem.measurement_df``.
        """

        # create rdata-like dicts from the prediction result
        @dataclass
        class FakeRData:
            ts: np.ndarray
            y: np.ndarray

        rdatas = [
            FakeRData(ts=condition.timepoints, y=condition.output)
            for condition in prediction.conditions
        ]

        # add an AMICI model, if possible
        model = None
        if predictor is not None:
            model = predictor.amici_objective.amici_model

        return self.rdatas_to_measurement_df(rdatas, model)

    def prediction_to_petab_simulation_df(
        self,
        prediction: PredictionResult,
        predictor: AmiciPredictor = None,
    ) -> pd.DataFrame:
        """
        See :meth:`prediction_to_petab_measurement_df`.

        Except a PEtab simulation dataframe is created, i.e. the measurement
        column label is adjusted.
        """
        return self.prediction_to_petab_measurement_df(
            prediction, predictor
        ).rename(columns={petab.MEASUREMENT: petab.SIMULATION})


class PetabSimulatorObjectiveCreator(ObjectiveCreator):
    """ObjectiveCreator for creating an objective based on a PEtabSimulator."""

    def __init__(
        self,
        petab_problem: petab.Problem,
        simulator: Simulator,
    ):
        self.petab_problem = petab_problem
        self.simulator = simulator

    def create_objective(self, **kwargs):
        """Create a PEtabSimulatorObjective."""
        return PetabSimulatorObjective(self.simulator)


class RoadRunnerObjectiveCreator(ObjectiveCreator):
    """ObjectiveCreator for creating an objective for a RoadRunner model."""

    def __init__(
        self,
        petab_problem: petab.Problem,
        rr: roadrunner.RoadRunner | None = None,
    ):
        self.petab_problem = petab_problem
        if rr is None:
            if roadrunner is None:
                raise ImportError(
                    "The `roadrunner` package (on PyPI: `libroadrunner`) "
                    "is required for this objective function."
                )
            rr = roadrunner.RoadRunner()
        self.rr = rr

    def _check_noise_formulae(
        self,
        edatas: list[ExpData] | None = None,
        parameter_mapping: list[ParMappingDictQuadruple] | None = None,
    ):
        """Check if the noise formulae are valid.

        Currently, only static values or singular parameters are supported.
        Complex formulae are not supported.
        """
        # check that parameter mapping is available
        if parameter_mapping is None:
            parameter_mapping = self.create_parameter_mapping()
        # check that edatas are available
        if edatas is None:
            edatas = self.create_edatas()
        # save formulae that need to be changed
        to_change = []
        # check that noise formulae are valid
        for i_edata, (edata, par_map) in enumerate(
            zip(edatas, parameter_mapping, strict=True)
        ):
            for j_formula, noise_formula in enumerate(edata.noise_formulae):
                # constant values are allowed
                if isinstance(noise_formula, numbers.Number):
                    continue
                # single parameters are allowed
                if noise_formula in par_map[1].keys():
                    continue
                # extract the observable name via regex pattern
                pattern = r"noiseParameter1_(.*?)($|\s)"
                observable_name = re.search(pattern, noise_formula).group(1)
                to_change.append((i_edata, j_formula, observable_name))
        # change formulae
        formulae_changed = []
        for i_edata, j_formula, obs_name in to_change:
            # assign new parameter, formula in RR and parameter into mapping
            original_formula = edatas[i_edata].noise_formulae[j_formula]
            edatas[i_edata].noise_formulae[j_formula] = (
                f"noiseFormula_{obs_name}"
            )
            # different conditions will have the same noise formula
            if (obs_name, original_formula) not in formulae_changed:
                self.rr.addParameter(f"noiseFormula_{obs_name}", 0.0, False)
                self.rr.addAssignmentRule(
                    f"noiseFormula_{obs_name}",
                    original_formula,
                    forceRegenerate=False,
                )
                self.rr.regenerateModel()
                formulae_changed.append((obs_name, original_formula))

    def _write_observables_to_model(self):
        """Write observables of petab problem to the model."""
        # add all observables as species
        for obs_id in self.petab_problem.observable_df.index:
            self.rr.addParameter(obs_id, 0.0, False)
        # extract all parameters from observable formulas
        parameters = petab.get_output_parameters(
            self.petab_problem.observable_df,
            self.petab_problem.model,
            noise=True,
            observables=True,
        )
        # add all parameters to the model
        for param_id in parameters:
            self.rr.addParameter(param_id, 0.0, False)
        formulae = self.petab_problem.observable_df[
            OBSERVABLE_FORMULA
        ].to_dict()

        # add all observable formulas as assignment rules
        for obs_id, formula in formulae.items():
            self.rr.addAssignmentRule(obs_id, formula, forceRegenerate=False)

        # regenerate model to apply changes
        self.rr.regenerateModel()

    def create_edatas(self) -> list[ExpData]:
        """Create a List of :class:`ExpData` objects from the PEtab problem."""
        # Create Dataframes per condition
        return ExpData.from_petab_problem(self.petab_problem)

    def fill_model(self):
        """Fill the RoadRunner model inplace from the PEtab problem.

        Parameters
        ----------
        return_model:
            Flag indicating if the model should be returned.
        """
        if not isinstance(self.petab_problem.model, SbmlModel):
            raise ValueError(
                "The model is not an SBML model. Using "
                "RoadRunner as simulator requires an SBML model."
            )  # TODO: add Pysb support
        if self.petab_problem.model.sbml_document:
            sbml_document = self.petab_problem.model.sbml_document
        elif self.petab_problem.model.sbml_model:
            sbml_document = (
                self.petab_problem.model.sbml_model.getSBMLDocument()
            )
        else:
            raise ValueError("No SBML model found.")
        sbml_writer = libsbml.SBMLWriter()
        sbml_string = sbml_writer.writeSBMLToString(sbml_document)
        self.rr.load(sbml_string)
        self._write_observables_to_model()

    def create_parameter_mapping(self):
        """Create a parameter mapping from the PEtab problem."""
        simulation_conditions = (
            self.petab_problem.get_simulation_conditions_from_measurement_df()
        )
        mapping = petab.get_optimization_to_simulation_parameter_mapping(
            condition_df=self.petab_problem.condition_df,
            measurement_df=self.petab_problem.measurement_df,
            parameter_df=self.petab_problem.parameter_df,
            observable_df=self.petab_problem.observable_df,
            model=self.petab_problem.model,
        )
        # check whether any species in the condition table are assigned
        species = self.rr.model.getFloatingSpeciesIds()
        # overrides in parameter table are handled already
        overrides = [
            specie
            for specie in species
            if specie in self.petab_problem.condition_df.columns
        ]
        if not overrides:
            return mapping
        for (_, condition), mapping_per_condition in zip(
            simulation_conditions.iterrows(), mapping, strict=True
        ):
            for override in overrides:
                preeq_id = condition.get(PREEQUILIBRATION_CONDITION_ID)
                sim_id = condition.get(SIMULATION_CONDITION_ID)
                if preeq_id:
                    parameter_id_or_value = (
                        self.petab_problem.condition_df.loc[preeq_id, override]
                    )
                    mapping_per_condition[0][override] = parameter_id_or_value
                    if isinstance(parameter_id_or_value, str):
                        mapping_per_condition[2][override] = (
                            self.petab_problem.parameter_df.loc[
                                parameter_id_or_value, petab.PARAMETER_SCALE
                            ]
                        )
                    elif isinstance(parameter_id_or_value, numbers.Number):
                        mapping_per_condition[2][override] = LIN
                    else:
                        raise ValueError(
                            "The parameter value in the condition table "
                            "is not a number or a parameter ID."
                        )
                if sim_id:
                    parameter_id_or_value = (
                        self.petab_problem.condition_df.loc[sim_id, override]
                    )
                    mapping_per_condition[1][override] = parameter_id_or_value
                    if isinstance(parameter_id_or_value, str):
                        mapping_per_condition[3][override] = (
                            self.petab_problem.parameter_df.loc[
                                parameter_id_or_value, petab.PARAMETER_SCALE
                            ]
                        )
                    elif isinstance(parameter_id_or_value, numbers.Number):
                        mapping_per_condition[3][override] = LIN
                    else:
                        raise ValueError(
                            "The parameter value in the condition table "
                            "is not a number or a parameter ID."
                        )
        return mapping

    def create_objective(
        self,
        rr: roadrunner.RoadRunner | None = None,
        edatas: ExpData | None = None,
    ) -> RoadRunnerObjective:
        """Create a :class:`pypesto.objective.RoadRunnerObjective`.

        Parameters
        ----------
        rr:
            RoadRunner instance.
        edatas:
            ExpData object.
        """
        roadrunner_instance = rr
        if roadrunner_instance is None:
            roadrunner_instance = self.rr
            self.fill_model()
        if edatas is None:
            edatas = self.create_edatas()

        parameter_mapping = self.create_parameter_mapping()

        # get x_names
        x_names = self.petab_problem.get_x_ids()

        calculator = RoadRunnerCalculator()

        # run the check for noise formulae
        self._check_noise_formulae(edatas, parameter_mapping)

        return RoadRunnerObjective(
            rr=roadrunner_instance,
            edatas=edatas,
            parameter_mapping=parameter_mapping,
            petab_problem=self.petab_problem,
            calculator=calculator,
            x_names=x_names,
            x_ids=x_names,
        )



"""Pymc v4 Sampler for Variational Inference."""

import logging
from typing import Optional

import numpy as np
import pytensor.tensor as pt
from scipy import stats

from ..objective import FD
from ..result import McmcPtResult
from ..sample.pymc import PymcObjectiveOp, PymcSampler
from ..sample.sampler import SamplerImportError

logger = logging.getLogger(__name__)


# implementation based on the pymc sampler code in pypesto and:
# https://www.pymc.io/projects/examples/en/latest/variational_inference/variational_api_quickstart.html


class PymcVariational(PymcSampler):
    """Wrapper around Pymc v4 variational inference.

    Parameters
    ----------
    step_function:
        A pymc step function, e.g. NUTS, Slice. If not specified, pymc
        determines one automatically (preferable).
    **kwargs:
        Options are directly passed on to `pymc.fit`.
    """

    def fit(
        self,
        n_iterations: int,
        method: str = "advi",
        random_seed: int | None = None,
        start_sigma: Optional = None,
        inf_kwargs: Optional = None,
        beta: float = 1.0,
        **kwargs,
    ):
        """
        Sample the problem.

        Parameters
        ----------
        n_iterations:
            Number of iterations.
        method: str or :class:`Inference` of pymc
            string name is case-insensitive in:
            -   'advi'  for ADVI
            -   'fullrank_advi'  for FullRankADVI
            -   'svgd'  for Stein Variational Gradient Descent
            -   'asvgd'  for Amortized Stein Variational Gradient Descent
        random_seed: int
            random seed for reproducibility
        start_sigma: `dict[str, np.ndarray]`
            starting standard deviation for inference, only available for method 'advi'
        inf_kwargs: dict
            additional kwargs passed to pymc.Inference
        beta:
            Inverse temperature (e.g. in parallel tempering).
        """
        try:
            import pymc
        except ImportError:
            raise SamplerImportError("pymc") from None

        problem = self.problem
        if not problem.objective.has_grad:
            logger.info(
                "The objective function does not provide gradients. "
                "Finite differences will be used."
            )
            problem.objective = FD(obj=problem.objective)
        log_post = PymcObjectiveOp.create_instance(problem.objective, beta)

        x0 = None
        x_names_free = problem.get_reduced_vector(problem.x_names)
        if self.x0 is not None:
            x0 = {
                x_name: val
                # FIXME: address https://github.com/ICB-DCM/pyPESTO/issues/1681
                #  and change to strict=True
                for x_name, val in zip(problem.x_names, self.x0, strict=False)
                if x_name in x_names_free
            }

        # create model context
        with pymc.Model():
            # parameter bounds as uniform prior
            _k = [
                pymc.Uniform(x_name, lower=lb, upper=ub)
                for x_name, lb, ub in zip(
                    x_names_free, problem.lb, problem.ub, strict=True
                )
            ]

            # convert parameters to PyTensor tensor variable
            theta = pt.as_tensor_variable(_k)

            # define distribution with log-posterior as density
            pymc.Potential("potential", log_post(theta))

            # record function values
            pymc.Deterministic("loggyposty", log_post(theta))

            # perform the actual sampling
            data = pymc.fit(
                n=int(n_iterations),
                method=method,
                random_seed=random_seed,
                start=x0,
                start_sigma=start_sigma,
                inf_kwargs=inf_kwargs,
                **kwargs,
            )

        self.data = data

    def sample(self, n_samples: int, beta: float = 1.0) -> McmcPtResult:
        """
        Sample from the variational approximation and return McmcPtResult object.

        Parameters
        ----------
        n_samples:
            Number of samples to be computed.
        """
        # get InferenceData object
        pymc_data = self.data.sample(n_samples)
        x_names_free = self.problem.get_reduced_vector(self.problem.x_names)
        post_samples = np.concatenate(
            [pymc_data.posterior[name].values for name in x_names_free]
        ).T
        return McmcPtResult(
            trace_x=post_samples[np.newaxis, :],
            trace_neglogpost=pymc_data.posterior.loggyposty.values,
            trace_neglogprior=np.full(
                pymc_data.posterior.loggyposty.values.shape, np.nan
            ),
            betas=np.array([1.0] * post_samples.shape[0]),
            burn_in=0,
            auto_correlation=0,
            effective_sample_size=n_samples,
            message="variational inference results",
        )

    def get_variational_parameters(self) -> (list, list):
        """Get the internal pymc variational parameters."""
        return (
            [param.name for param in self.data.params],
            [param.eval() for param in self.data.params],
        )

    def set_variational_parameters(self, param_list: list):
        """
        Set the internal pymc variational parameters.

        Parameters
        ----------
        param_list:
            List of tuples of the form (param_name, param_value).
        """
        if len(param_list) != len(self.data.params):
            raise ValueError(
                "The number of parameters does not match the number of variational parameters."
            )
        for i, param in enumerate(param_list):
            self.data.params[i].set_value(param)

    def eval_variational_log_density(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate the log density of the variational approximation at x_points.

        Parameters
        ----------
        x:
            The points at which to evaluate the log density.
        """
        # TODO: add support for other methods
        logger.warning(
            "currently only supports the methods `advi` and `fullrank_advi`"
        )

        if x.ndim == 1:
            x = x.reshape(1, -1)
        log_density_at_points = np.zeros_like(x)
        for i, point in enumerate(x):
            log_density_at_points[i] = stats.multivariate_normal.logpdf(
                point, mean=self.data.mean.eval(), cov=self.data.cov.eval()
            )
        vi_log_density = np.sum(log_density_at_points, axis=-1)
        return vi_log_density



"""Contains the PetabImporter class."""

from __future__ import annotations

import logging
import os
import tempfile
import warnings
from collections.abc import Callable, Iterable, Sequence
from importlib.metadata import version
from typing import (
    Any,
)

import pandas as pd
import petab.v1 as petab

try:
    import roadrunner
except ImportError:
    roadrunner = None

from ..C import (
    AMICI,
    CENSORED,
    ORDINAL,
    ORDINAL_OPTIONS,
    PETAB,
    ROADRUNNER,
    SEMIQUANTITATIVE,
    SPLINE_APPROXIMATION_OPTIONS,
)
from ..hierarchical.inner_calculator_collector import InnerCalculatorCollector
from ..objective import AggregatedObjective, AmiciObjective, ObjectiveBase
from ..objective.priors import NegLogParameterPriors, get_parameter_prior_dict
from ..predict import AmiciPredictor
from ..problem import HierarchicalProblem, Problem
from ..result import PredictionResult
from ..startpoint import StartpointMethod
from .objective_creator import (
    AmiciObjectiveCreator,
    ObjectiveCreator,
    PetabSimulatorObjectiveCreator,
    RoadRunnerObjectiveCreator,
)
from .util import PetabStartpoints, get_petab_non_quantitative_data_types

try:
    import amici
except ImportError:
    amici = None

logger = logging.getLogger(__name__)


class PetabImporter:
    """
    Importer for PEtab files.

    Create an :class:`amici.amici.Model`, an :class:`pypesto.objective.AmiciObjective` or a
    :class:`pypesto.problem.Problem` from PEtab files. The created objective function is a
    negative log-likelihood function and can thus be negative. The actual
    form of the likelihood depends on the noise model specified in the provided PEtab problem.
    For more information, see the
    `PEtab documentation <https://petab.readthedocs.io/en/latest/documentation_data_format.html#noise-distributions>`_.
    """  # noqa

    MODEL_BASE_DIR = f"amici_models/{version('amici') if amici else ''}"

    def __init__(
        self,
        petab_problem: petab.Problem,
        output_folder: str | None = None,
        model_name: str | None = None,
        validate_petab: bool = True,
        validate_petab_hierarchical: bool = True,
        hierarchical: bool = False,
        inner_options: dict | None = None,
        simulator_type: str = AMICI,
        simulator: petab.Simulator | None = None,
        rr: roadrunner.RoadRunner | None = None,
    ):
        """Initialize importer.

        Parameters
        ----------
        petab_problem:
            Managing access to the model and data.
        output_folder:
            Folder to contain the amici model. Defaults to
            './amici_models/{model_name}'.
        model_name:
            Name of the model, which will in particular be the name of the
            compiled model python module.
        validate_petab:
            Flag indicating if the PEtab problem shall be validated.
        validate_petab_hierarchical:
            Flag indicating if the PEtab problem shall be validated in terms of
            pyPESTO's hierarchical optimization implementation.
        hierarchical:
            Whether to use hierarchical optimization or not, in case the
            underlying PEtab problem has parameters marked for hierarchical
            optimization (non-empty `parameterType` column in the PEtab
            parameter table). Required for ordinal, censored and semiquantitative data.
        inner_options:
            Options for the inner problems and solvers.
            If not provided, default options will be used.
        simulator_type:
            The type of simulator to use. Depending on this different kinds
            of objectives will be created. Allowed types are 'amici', 'petab',
            and 'roadrunner'.
        simulator:
            In case of a ``simulator_type == 'petab'``, the simulator object
            has to be provided. Otherwise, the argument is not used.
        """
        self.petab_problem = petab_problem
        self._hierarchical = hierarchical

        self._non_quantitative_data_types = (
            get_petab_non_quantitative_data_types(petab_problem)
        )

        if self._non_quantitative_data_types is None and hierarchical:
            raise ValueError(
                "Hierarchical optimization enabled, but no non-quantitative "
                "data types specified. Specify non-quantitative data types "
                "or disable hierarchical optimization."
            )

        if (
            self._non_quantitative_data_types is not None
            and any(
                data_type in self._non_quantitative_data_types
                for data_type in [ORDINAL, CENSORED, SEMIQUANTITATIVE]
            )
            and not self._hierarchical
        ):
            raise ValueError(
                "Ordinal, censored and semiquantitative data require "
                "hierarchical optimization to be enabled.",
            )

        self.inner_options = inner_options
        if self.inner_options is None:
            self.inner_options = {}

        self.validate_inner_options()

        self.validate_petab = validate_petab
        if self.validate_petab:
            if petab.lint_problem(petab_problem):
                raise ValueError("Invalid PEtab problem.")
        if self._hierarchical and validate_petab_hierarchical:
            from ..hierarchical.petab import (
                validate_hierarchical_petab_problem,
            )

            validate_hierarchical_petab_problem(petab_problem)

        if output_folder is None:
            output_folder = _find_output_folder_name(
                self.petab_problem,
                model_name=model_name,
            )
        self.output_folder = output_folder

        if model_name is None:
            model_name = _find_model_name(self.output_folder)
        self.model_name = model_name

        self.simulator_type = simulator_type
        self.simulator = simulator
        if simulator_type == PETAB and simulator is None:
            raise ValueError(
                "A petab simulator object must be provided if the simulator "
                "type is 'petab'."
            )
        self.roadrunner_instance = rr

    @staticmethod
    def from_yaml(
        yaml_config: dict | str,
        output_folder: str = None,
        model_name: str = None,
        simulator_type: str = AMICI,
    ) -> PetabImporter:
        """Simplified constructor using a petab yaml file."""
        petab_problem = petab.Problem.from_yaml(yaml_config)

        return PetabImporter(
            petab_problem=petab_problem,
            output_folder=output_folder,
            model_name=model_name,
            simulator_type=simulator_type,
        )

    def validate_inner_options(self):
        """Validate the inner options."""
        for key in self.inner_options:
            if key not in ORDINAL_OPTIONS + SPLINE_APPROXIMATION_OPTIONS:
                raise ValueError(f"Unknown inner option {key}.")

    def check_gradients(
        self,
        *args,
        rtol: float = 1e-2,
        atol: float = 1e-3,
        mode: str | list[str] = None,
        multi_eps=None,
        **kwargs,
    ) -> bool:
        """
        Check if gradients match finite differences (FDs).

        Parameters
        ----------
        rtol: relative error tolerance
        atol: absolute error tolerance
        mode: function values or residuals
        objAbsoluteTolerance: absolute tolerance in sensitivity calculation
        objRelativeTolerance: relative tolerance in sensitivity calculation
        multi_eps: multiple test step width for FDs

        Returns
        -------
        match: Whether gradients match FDs (True) or not (False)
        """
        raise NotImplementedError(
            "This function has been removed. "
            "Please use `objective.check_gradients_match_finite_differences`."
        )

    def create_prior(self) -> NegLogParameterPriors | None:
        """
        Create a prior from the parameter table.

        Returns None, if no priors are defined.
        """
        prior_list = []

        if petab.OBJECTIVE_PRIOR_TYPE in self.petab_problem.parameter_df:
            for i, x_id in enumerate(self.petab_problem.x_ids):
                prior_type_entry = self.petab_problem.parameter_df.loc[
                    x_id, petab.OBJECTIVE_PRIOR_TYPE
                ]

                if (
                    isinstance(prior_type_entry, str)
                    and prior_type_entry != petab.PARAMETER_SCALE_UNIFORM
                ):
                    # check if parameter for which prior is defined is a fixed parameter
                    if x_id in self.petab_problem.x_fixed_ids:
                        logger.warning(
                            f"Parameter {x_id} is marked as fixed but has a "
                            f"prior defined. This might be unintended."
                        )

                    prior_params = [
                        float(param)
                        for param in self.petab_problem.parameter_df.loc[
                            x_id, petab.OBJECTIVE_PRIOR_PARAMETERS
                        ].split(";")
                    ]

                    scale = self.petab_problem.parameter_df.loc[
                        x_id, petab.PARAMETER_SCALE
                    ]

                    prior_list.append(
                        get_parameter_prior_dict(
                            i, prior_type_entry, prior_params, scale
                        )
                    )

        if len(prior_list):
            return NegLogParameterPriors(prior_list)
        else:
            return None

    def create_startpoint_method(self, **kwargs) -> StartpointMethod:
        """Create a startpoint method.

        Parameters
        ----------
        **kwargs:
            Additional keyword arguments passed on to
            :meth:`pypesto.startpoint.FunctionStartpoints.__init__`.
        """
        return PetabStartpoints(petab_problem=self.petab_problem, **kwargs)

    def create_objective_creator(
        self,
        simulator_type: str = AMICI,
        simulator: petab.Simulator | None = None,
    ) -> ObjectiveCreator:
        """Choose :class:`ObjectiveCreator` depending on the simulator type.

        Parameters
        ----------
        simulator_type:
            The type of simulator to use. Depending on this different kinds
            of objectives will be created. Allowed types are 'amici', 'petab',
            and 'roadrunner'.
        simulator:
            In case of a ``simulator_type == 'petab'``, the simulator object
            has to be provided. Otherwise the argument is not used.

        """
        if simulator_type == AMICI:
            return AmiciObjectiveCreator(
                petab_problem=self.petab_problem,
                output_folder=self.output_folder,
                model_name=self.model_name,
                hierarchical=self._hierarchical,
                inner_options=self.inner_options,
                non_quantitative_data_types=self._non_quantitative_data_types,
                validate_petab=self.validate_petab,
            )
        elif simulator_type == PETAB:
            return PetabSimulatorObjectiveCreator(
                petab_problem=self.petab_problem, simulator=simulator
            )
        elif simulator_type == ROADRUNNER:
            return RoadRunnerObjectiveCreator(
                petab_problem=self.petab_problem, rr=self.roadrunner_instance
            )

    def create_problem(
        self,
        objective: ObjectiveBase = None,
        x_guesses: Iterable[float] | None = None,
        problem_kwargs: dict[str, Any] = None,
        startpoint_kwargs: dict[str, Any] = None,
        **kwargs,
    ) -> Problem:
        """Create a :class:`pypesto.problem.Problem`.

        Parameters
        ----------
        objective:
            Objective as created by :meth:`create_objective`.
        x_guesses:
            Guesses for the parameter values, shape (g, dim), where g denotes
            the number of guesses. These are used as start points in the
            optimization.
        problem_kwargs:
            Passed to :meth:`pypesto.problem.Problem.__init__`.
        startpoint_kwargs:
            Keyword arguments forwarded to
            :meth:`PetabImporter.create_startpoint_method`.
        **kwargs:
            Additional key word arguments passed on to the objective,
            if not provided.

        Returns
        -------
        A :class:`pypesto.problem.Problem` for the objective.
        """
        if objective is None:
            self.objective_constructor = self.create_objective_creator(
                kwargs.pop("simulator_type", self.simulator_type),
                kwargs.pop("simulator", self.simulator),
            )
            objective = self.objective_constructor.create_objective(**kwargs)

        x_fixed_indices = self.petab_problem.x_fixed_indices
        x_fixed_vals = self.petab_problem.x_nominal_fixed_scaled
        x_ids = self.petab_problem.x_ids
        lb = self.petab_problem.lb_scaled
        ub = self.petab_problem.ub_scaled

        # Raise error if the correct calculator is not used.
        if self._hierarchical:
            if not isinstance(objective.calculator, InnerCalculatorCollector):
                raise AssertionError(
                    f"If hierarchical optimization is enabled, the `calculator` attribute of the `objective` has to be {InnerCalculatorCollector} and not {objective.calculator}."
                )

        # In case of hierarchical optimization, parameters estimated in the
        # inner subproblem are removed from the outer problem
        if self._hierarchical:
            inner_parameter_ids = objective.calculator.get_inner_par_ids()
            lb = [
                b
                for x, b in zip(x_ids, lb, strict=True)
                if x not in inner_parameter_ids
            ]
            ub = [
                b
                for x, b in zip(x_ids, ub, strict=True)
                if x not in inner_parameter_ids
            ]
            x_ids = [x for x in x_ids if x not in inner_parameter_ids]
            x_fixed_indices = list(
                map(x_ids.index, self.petab_problem.x_fixed_ids)
            )

        x_scales = [
            self.petab_problem.parameter_df.loc[x_id, petab.PARAMETER_SCALE]
            for x_id in x_ids
        ]

        if problem_kwargs is None:
            problem_kwargs = {}

        if startpoint_kwargs is None:
            startpoint_kwargs = {}

        prior = self.create_prior()

        if prior is not None:
            if self._hierarchical:
                raise NotImplementedError(
                    "Hierarchical optimization in combination with priors "
                    "is not yet supported."
                )
            objective = AggregatedObjective([objective, prior])
            logger.info("Created an aggregated objective using priors.")

        if self._hierarchical:
            problem_class = HierarchicalProblem
        else:
            problem_class = Problem

        problem = problem_class(
            objective=objective,
            lb=lb,
            ub=ub,
            x_fixed_indices=x_fixed_indices,
            x_fixed_vals=x_fixed_vals,
            x_guesses=x_guesses,
            x_names=x_ids,
            x_scales=x_scales,
            x_priors_defs=prior,
            startpoint_method=self.create_startpoint_method(
                **startpoint_kwargs
            ),
            **problem_kwargs,
        )

        return problem

    def create_model(
        self,
        force_compile: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> amici.Model:
        """See :meth:`AmiciObjectiveCreator.create_model`."""
        warnings.warn(
            "This function has been moved to `AmiciObjectiveCreator`.",
            DeprecationWarning,
            stacklevel=2,
        )
        objective_constructor = self.create_objective_creator(
            kwargs.pop("simulator_type", self.simulator_type),
            kwargs.pop("simulator", self.simulator),
        )
        return objective_constructor.create_model(
            force_compile=force_compile,
            verbose=verbose,
            **kwargs,
        )

    def create_objective(
        self,
        model: amici.Model = None,
        solver: amici.Solver = None,
        edatas: Sequence[amici.ExpData] = None,
        force_compile: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> ObjectiveBase:
        """See :meth:`AmiciObjectiveCreator.create_objective`."""
        warnings.warn(
            "This function has been moved to `AmiciObjectiveCreator`.",
            DeprecationWarning,
            stacklevel=2,
        )
        objective_constructor = self.create_objective_creator(
            kwargs.pop("simulator_type", self.simulator_type),
            kwargs.pop("simulator", self.simulator),
        )
        return objective_constructor.create_objective(
            model=model,
            solver=solver,
            edatas=edatas,
            force_compile=force_compile,
            verbose=verbose,
            **kwargs,
        )

    def create_predictor(
        self,
        objective: AmiciObjective = None,
        amici_output_fields: Sequence[str] = None,
        post_processor: Callable | None = None,
        post_processor_sensi: Callable | None = None,
        post_processor_time: Callable | None = None,
        max_chunk_size: int | None = None,
        output_ids: Sequence[str] = None,
        condition_ids: Sequence[str] = None,
    ) -> AmiciPredictor:
        """See :meth:`AmiciObjectiveCreator.create_predictor`."""
        if self.simulator_type != AMICI:
            raise ValueError(
                "Predictor can only be created for amici models and is "
                "supposed to be created from the AmiciObjectiveCreator."
            )
        warnings.warn(
            "This function has been moved to `AmiciObjectiveCreator`.",
            DeprecationWarning,
            stacklevel=2,
        )
        objective_constructor = self.create_objective_creator()
        return objective_constructor.create_predictor(
            objective=objective,
            amici_output_fields=amici_output_fields,
            post_processor=post_processor,
            post_processor_sensi=post_processor_sensi,
            post_processor_time=post_processor_time,
            max_chunk_size=max_chunk_size,
            output_ids=output_ids,
            condition_ids=condition_ids,
        )

    def rdatas_to_measurement_df(
        self,
        rdatas: Sequence[amici.ReturnData],
        model: amici.Model = None,
        verbose: bool = True,
    ) -> pd.DataFrame:
        """See :meth:`AmiciObjectiveCreator.rdatas_to_measurement_df`."""
        raise NotImplementedError(
            "This function has been moved to `AmiciObjectiveCreator`."
        )

    def rdatas_to_simulation_df(
        self,
        rdatas: Sequence[amici.ReturnData],
        model: amici.Model = None,
    ) -> pd.DataFrame:
        """
        See :meth:`rdatas_to_measurement_df`.

        Except a petab simulation dataframe is created, i.e. the measurement
        column label is adjusted.
        """
        raise NotImplementedError(
            "This function has been moved to `AmiciObjectiveCreator`."
        )

    def prediction_to_petab_measurement_df(
        self,
        prediction: PredictionResult,
        predictor: AmiciPredictor = None,
    ) -> pd.DataFrame:
        """
        Cast prediction into a dataframe.

        If a PEtab problem is simulated without post-processing, then the
        result can be cast into a PEtab measurement or simulation dataframe

        Parameters
        ----------
        prediction:
            A prediction result as produced by an :class:`pypesto.predict.AmiciPredictor`.
        predictor:
            The :class:`pypesto.predict.AmiciPredictor` instance.

        Returns
        -------
        A dataframe built from the rdatas in the format as in
        ``self.petab_problem.measurement_df``.
        """
        raise NotImplementedError(
            "This function has been moved to `AmiciObjectiveCreator`."
        )

    def prediction_to_petab_simulation_df(
        self,
        prediction: PredictionResult,
        predictor: AmiciPredictor = None,
    ) -> pd.DataFrame:
        """
        See :meth:`prediction_to_petab_measurement_df`.

        Except a PEtab simulation dataframe is created, i.e. the measurement
        column label is adjusted.
        """
        raise NotImplementedError(
            "This function has been moved to `AmiciObjectiveCreator`."
        )


def _find_output_folder_name(
    petab_problem: petab.Problem,
    model_name: str,
) -> str:
    """
    Find a name for storing the compiled amici model in.

    If available, use the model name from the ``petab_problem`` or the
    provided ``model_name`` (latter is given priority), otherwise create a
    unique name. The folder will be located in the
    :obj:`PetabImporter.MODEL_BASE_DIR` subdirectory of the current directory.
    """
    # check whether location for amici model is a file
    if os.path.exists(PetabImporter.MODEL_BASE_DIR) and not os.path.isdir(
        PetabImporter.MODEL_BASE_DIR
    ):
        raise AssertionError(
            f"{PetabImporter.MODEL_BASE_DIR} exists and is not a directory, "
            f"thus cannot create a directory for the compiled amici model."
        )

    # create base directory if non-existent
    if not os.path.exists(PetabImporter.MODEL_BASE_DIR):
        os.makedirs(PetabImporter.MODEL_BASE_DIR)

    # try model id
    model_id = petab_problem.model.model_id
    if model_name is not None:
        model_id = model_name

    if model_id:
        output_folder = os.path.abspath(
            os.path.join(PetabImporter.MODEL_BASE_DIR, model_id)
        )
    else:
        # create random folder name
        output_folder = os.path.abspath(
            tempfile.mkdtemp(dir=PetabImporter.MODEL_BASE_DIR)
        )
    return output_folder


def _find_model_name(output_folder: str) -> str:
    """Just re-use the last part of the output folder."""
    return os.path.split(os.path.normpath(output_folder))[-1]



import matplotlib.cm as cm
import numpy as np
from matplotlib.colors import is_color_like

from pypesto.util import assign_clusters

# for typehints
from ..C import COLOR


def assign_clustered_colors(
    vals: np.ndarray, balance_alpha: bool = True, highlight_global: bool = True
):
    """
    Cluster and assign colors.

    Parameters
    ----------
    vals:
        List to be clustered and assigned colors.
    balance_alpha:
        Flag indicating whether alpha for large clusters should be reduced to
        avoid overplotting
    highlight_global:
        flag indicating whether global optimum should be highlighted

    Returns
    -------
    colors: list of RGBA
        One for each element in 'vals'.
    """
    # sanity checks
    if vals is None or len(vals) == 0:
        return []

    # assign clusters
    clusters, cluster_size = assign_clusters(vals)

    # create list of colors, which has the correct shape
    n_clusters = 1 + max(clusters) - sum(cluster_size == 1)

    # if best value was found more than once: we need one color less
    if highlight_global and cluster_size[0] > 1:
        n_clusters -= 1

    # fill color array from colormap
    colormap = cm.ScalarMappable().to_rgba
    color_list = colormap(np.linspace(0.0, 1.0, n_clusters))

    # best optimum should be colored in red
    if highlight_global and cluster_size[0] > 1:
        color_list = np.concatenate(([[1.0, 0.0, 0.0, 1.0]], color_list))

    # We have clustered the results. However, clusters may have size 1,
    # so we need to rearrange the regroup the results into "no_clusters",
    # which will be grey, and "real_clusters", which will be colored

    # run through the cluster_indices and collect which element
    # is in 1-element-clusters, and which is in real clusters
    no_clusters = np.where(cluster_size == 1)[0]
    real_clusters = np.unique(np.where(cluster_size > 1)[0])

    # assign transparency values according to cluster size, if wanted
    if balance_alpha:
        # set minimal alpha value to avoid non-visible colors
        min_alpha = 0.01
        # assign neutral color, add 1 for avoiding division by zero
        grey = [0.7, 0.7, 0.7, min(1.0, 5.0 / (no_clusters.size + 1.0))]

        # reduce alpha level depend on size of each cluster
        n_cluster_size = np.delete(cluster_size, no_clusters)
        for icluster in range(n_clusters):
            color_list[icluster][3] = min(
                1.0, max(5.0 / n_cluster_size[icluster], min_alpha)
            )
    else:
        # assign neutral color
        grey = [0.7, 0.7, 0.7, 1.0]

    # create a color list, prfilled with grey values
    colors = np.array([grey] * clusters.size)

    # assign colors to real clusters
    for icol, iclust in enumerate(real_clusters):
        # find indices belonging to the cluster iclust and assign color
        ind_of_iclust = np.argwhere(clusters == iclust).flatten()
        colors[ind_of_iclust, :] = color_list[icol, :]

    # if best value was found only once: replace it with red
    if highlight_global and cluster_size[0] == 1:
        colors[0] = [1.0, 0.0, 0.0, 1.0]

    return colors


def assign_colors(
    vals: np.ndarray,
    colors: COLOR | list[COLOR] | np.ndarray | None = None,
    balance_alpha: bool = True,
    highlight_global: bool = True,
) -> np.ndarray:
    """
    Assign colors or format user specified colors.

    Parameters
    ----------
    vals:
        List to be clustered and assigned colors.
    colors:
        list of colors recognized by matplotlib, or single color
    balance_alpha:
        Flag indicating whether alpha for large clusters should be reduced to
        avoid overplotting
    highlight_global:
        flag indicating whether global optimum should be highlighted

    Returns
    -------
    colors: list of colors recognized by matplotlib
        One for each element in 'vals'.
    """
    # sanity checks
    if vals is None or len(vals) == 0:
        return np.array([])

    # if the user did not specify any colors:
    if colors is None:
        return assign_clustered_colors(
            vals,
            balance_alpha=balance_alpha,
            highlight_global=highlight_global,
        )

    # Get number of elements and use user assigned colors
    n_vals = len(vals) if isinstance(vals, list) else vals.size

    if is_color_like(colors):
        # one color was passed
        return np.array([colors] * n_vals)

    elif (
        isinstance(colors, (list, np.ndarray))
        and len(colors) == len(vals)
        and sum([is_color_like(c) for c in colors]) == len(colors)
    ):
        # a list of colors was passed, one for each value in vals

        # convert to ndarray
        colors = np.array(colors)
        return colors

    # Shape of array did not match n_vals. Error due to size mismatch:
    raise ValueError(
        "Incorrect color input. Colors must be specified either as "
        "list of colors recognized by matplotlib with length equal to that of `vals` "
        f"(here: {n_vals}), or as a single matplotlib color."
    )


def assign_colors_for_list(
    num_entries: int,
    colors: COLOR | list[COLOR] | np.ndarray | None = None,
) -> list[list[float]] | np.ndarray:
    """
    Create a list of colors for a list of items.

    Can also check a user-provided list of colors and use this if
    everything is ok.

    Parameters
    ----------
    num_entries:
        number of results in list
    colors:
        list of colors, or single color

    Returns
    -------
    colors:
        List of RGBA, one for each element in 'vals'.
    """
    # if the user did not specify any colors:
    if colors is None:
        # default colors will be used, one for each entry in the result list.
        # Colors are created from assign_colors, which needs a dummy list
        # doubled for clustering
        dummy_clusters = np.array(list(range(num_entries)) * 2)

        # we don't want alpha levels for all plotting routines in this case...
        colors = assign_colors(
            dummy_clusters, balance_alpha=False, highlight_global=False
        )

        # dummy cluster had twice as many entries as really there. Reduce.
        real_indices = np.arange(0, colors.shape[0], 2)
        return colors[real_indices]

    # Pass the colors through assign_colors to check correct format of RGBA
    return assign_colors(
        vals=np.array(list(range(num_entries))),
        colors=colors,
        balance_alpha=False,
        highlight_global=False,
    )



"""Include functions for saving various results to hdf5."""

from __future__ import annotations

import logging
import os
from numbers import Integral
from pathlib import Path

import h5py
import numpy as np

from .. import OptimizeResult, OptimizerResult
from ..result import ProfilerResult, Result, SampleResult
from .hdf5 import write_array, write_float_array

logger = logging.getLogger(__name__)


def check_overwrite(f: h5py.File | h5py.Group, overwrite: bool, target: str):
    """
    Check whether target already exists.

    Sends a warning if ``overwrite=False``, deletes the target if
    ``overwrite=True``.

    Attributes
    ----------
    f: file or group where existence of a group with the path group_path
       should be checked
    target: name of the group, whose existence is checked
    overwrite: if ``True``, it deletes the target in ``f``
    """
    if target in f:
        if overwrite:
            del f[target]
        else:
            raise RuntimeError(
                f"File `{f.file.filename}` already exists and contains "
                f"information about {target} result. "
                f"If you wish to overwrite the file, set "
                f"`overwrite=True`."
            )


class ProblemHDF5Writer:
    """
    Writer of the HDF5 problem files.

    Attributes
    ----------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str | Path):
        """
        Initialize writer.

        Parameters
        ----------
        storage_filename:
            HDF5 problem file name
        """
        self.storage_filename = str(storage_filename)

    def write(self, problem, overwrite: bool = False):
        """Write HDF5 problem file from pyPESTO problem object."""
        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, "problem")
            attrs_to_save = [
                a
                for a in dir(problem)
                if not a.startswith("__")
                and not callable(getattr(problem, a))
                and not hasattr(type(problem), a)
            ]

            problem_grp = f.create_group("problem")
            # save the configuration
            f["problem/config"] = str(problem.objective.get_config())

            for problem_attr in attrs_to_save:
                value = getattr(problem, problem_attr)
                if isinstance(value, (list, np.ndarray)):
                    value = np.asarray(value)
                    if value.size:
                        write_array(problem_grp, problem_attr, value)
                elif isinstance(value, (Integral, str)):
                    problem_grp.attrs[problem_attr] = value


class OptimizationResultHDF5Writer:
    """
    Writer of the HDF5 result files.

    Attributes
    ----------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str | Path):
        """
        Initialize Writer.

        Parameters
        ----------
        storage_filename:
            HDF5 result file name
        """
        self.storage_filename = str(storage_filename)

    def write(
        self,
        result: Result
        | OptimizeResult
        | OptimizerResult
        | list[OptimizerResult],
        overwrite=False,
    ):
        """Write HDF5 result file from pyPESTO result object.

        Parameters
        ----------
        result: Result to be saved.
        overwrite: Boolean, whether already existing results should be
            overwritten. This applies to the whole list of results, not only to
            individual results. See :meth:`write_optimizer_result` for
            incrementally writing a sequence of `OptimizerResult`.
        """
        Path(self.storage_filename).parent.mkdir(parents=True, exist_ok=True)

        if isinstance(result, Result):
            results = result.optimize_result.list
        elif isinstance(result, OptimizeResult):
            results = result.list
        elif isinstance(result, list):
            results = result
        elif isinstance(result, OptimizerResult):
            results = [result]
        else:
            raise ValueError(f"Unsupported type for `result`: {type(result)}.")

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, "optimization")
            optimization_grp = f.require_group("optimization")
            results_grp = optimization_grp.require_group("results")

            for start in results:
                self._do_write_optimizer_result(start, results_grp, overwrite)

    def write_optimizer_result(
        self, result: OptimizerResult, overwrite: bool = False
    ):
        """Write HDF5 result file from pyPESTO result object.

        Parameters
        ----------
        result: Result to be saved.
        overwrite: Boolean, whether already existing results with the same ID
            should be overwritten.s
        """
        Path(self.storage_filename).parent.mkdir(parents=True, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            results_grp = f.require_group("optimization/results")
            self._do_write_optimizer_result(result, results_grp, overwrite)

    def _do_write_optimizer_result(
        self, result: OptimizerResult, g: h5py.Group = None, overwrite=False
    ):
        """Write an OptimizerResult to the given group."""
        sub_group_id = result["id"]
        check_overwrite(g, overwrite, sub_group_id)
        start_grp = g.require_group(sub_group_id)
        for key in result.keys():
            if key == "history":
                continue
            if isinstance(result[key], np.ndarray):
                write_array(start_grp, key, result[key])
            elif result[key] is not None:
                start_grp.attrs[key] = result[key]


class SamplingResultHDF5Writer:
    """
    Writer of the HDF5 sampling files.

    Attributes
    ----------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str | Path):
        """
        Initialize Writer.

        Parameters
        ----------
        storage_filename:
            HDF5 result file name
        """
        self.storage_filename = str(storage_filename)

    def write(self, result: Result, overwrite: bool = False):
        """Write HDF5 sampling file from pyPESTO result object."""
        # if there is no sample available, log a warning and return
        # SampleResult is only a dummy class created by the Result class
        # and always indicates the lack of a sampling result.
        if isinstance(result.sample_result, SampleResult):
            logger.warning(
                "Warning: There is no sampling_result, "
                "which you tried to save to hdf5."
            )
            return

        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, "sampling")
            results_grp = f.require_group("sampling/results")

            for key in result.sample_result.keys():
                if isinstance(result.sample_result[key], np.ndarray):
                    write_float_array(
                        results_grp, key, result.sample_result[key]
                    )
                elif result.sample_result[key] is not None:
                    results_grp.attrs[key] = result.sample_result[key]
            f.flush()


class ProfileResultHDF5Writer:
    """
    Writer of the HDF5 result files.

    Attributes
    ----------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str | Path):
        """
        Initialize Writer.

        Parameters
        ----------
        storage_filename:
            HDF5 result file name
        """
        self.storage_filename = str(storage_filename)

    def write(self, result: Result, overwrite: bool = False):
        """Write HDF5 result file from pyPESTO result object."""
        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, "profiling")
            profiling_grp = f.require_group("profiling")

            for profile_id, profile in enumerate(result.profile_result.list):
                profile_grp = profiling_grp.require_group(str(profile_id))
                for parameter_id, parameter_profile in enumerate(profile):
                    result_grp = profile_grp.require_group(str(parameter_id))
                    self._write_profiler_result(parameter_profile, result_grp)

            f.flush()

    @staticmethod
    def _write_profiler_result(
        parameter_profile: ProfilerResult | None, result_grp: h5py.Group
    ) -> None:
        """Write a single ProfilerResult to hdf5.

        Writes a single profile for a single parameter to the provided HDF5 group.
        """
        if parameter_profile is None:
            result_grp.attrs["IsNone"] = True
            return

        result_grp.attrs["IsNone"] = False

        for key, value in parameter_profile.items():
            try:
                if isinstance(value, np.ndarray):
                    write_float_array(result_grp, key, value)
                elif value is not None:
                    result_grp.attrs[key] = value
            except Exception as e:
                raise ValueError(
                    f"Error writing {key} ({value}) to {result_grp}."
                ) from e


def write_result(
    result: Result,
    filename: str | Path,
    overwrite: bool = False,
    problem: bool = True,
    optimize: bool = False,
    profile: bool = False,
    sample: bool = False,
):
    """
    Save whole pypesto.Result to hdf5 file.

    Boolean indicators allow specifying what to save.

    Parameters
    ----------
    result:
        The :class:`pypesto.Result` object to be saved.
    filename:
        The HDF5 filename.
    overwrite:
        Boolean, whether already existing results should be overwritten.
    problem:
        Read the problem.
    optimize:
        Read the optimize result.
    profile:
        Read the profile result.
    sample:
        Read the sample result.
    """
    if not any([optimize, profile, sample]):
        optimize = True
        profile = True
        sample = True

    if problem:
        pypesto_problem_writer = ProblemHDF5Writer(filename)
        pypesto_problem_writer.write(result.problem, overwrite=overwrite)

    if optimize:
        pypesto_opt_writer = OptimizationResultHDF5Writer(filename)
        pypesto_opt_writer.write(result, overwrite=overwrite)

    if profile:
        pypesto_profile_writer = ProfileResultHDF5Writer(filename)
        pypesto_profile_writer.write(result, overwrite=overwrite)

    if sample:
        pypesto_sample_writer = SamplingResultHDF5Writer(filename)
        pypesto_sample_writer.write(result, overwrite=overwrite)

    if hasattr(result, "variational_result"):
        logger.warning(
            "Results from variational inference are not saved in the hdf5 file. "
            "You have to save them manually."
        )



"""Interface to PEtab.jl."""

import logging
import os

import numpy as np

from .base import JuliaObjective, _read_source

logger = logging.getLogger(__name__)

PEtabProblemJl = ["PEtab.jl::PEtabODEProblem"]


class PEtabJlObjective(JuliaObjective):
    """
    Wrapper around an objective defined in PEtab.jl.

    Parameters
    ----------
    module:
        Name of the julia module containing the objective.
    source_file:
        Julia source file. Defaults to "{module}.jl".
    petab_problem_name:
        Name of the petab problem variable in the julia module.
    """

    def __init__(
        self,
        module: str,
        source_file: str = None,
        petab_problem_name: str = "petabProblem",
        precompile: bool = True,
        force_compile: bool = False,
    ):
        """Initialize objective."""
        # lazy imports
        try:
            from julia import Main, Pkg  # noqa: F401

            Pkg.activate(".")
        except ImportError:
            raise ImportError(
                "Install PyJulia, e.g. via `pip install pypesto[julia]`, "
                "and see the class documentation",
            ) from None

        self.module = module
        self.source_file = source_file
        self._petab_problem_name = petab_problem_name
        if precompile:
            self.precompile_model(force_compile=force_compile)

        if self.source_file is None:
            self.source_file = f"{module}.jl"

        # Include module if not already included
        _read_source(module, source_file)

        petab_jl_problem = self.get(petab_problem_name)
        self.petab_jl_problem = petab_jl_problem

        # get functions
        fun = self.petab_jl_problem.nllh
        grad = self.petab_jl_problem.grad
        hess = self.petab_jl_problem.hess
        x_names = np.asarray(self.petab_jl_problem.xnames)

        # call the super super super constructor
        super(JuliaObjective, self).__init__(
            fun=fun, grad=grad, hess=hess, x_names=x_names
        )

    def __getstate__(self):
        """Get state for pickling."""
        # if not dumped, dump it via JLD2
        return {
            "module": self.module,
            "source_file": self.source_file,
            "_petab_problem_name": self._petab_problem_name,
        }

    def __setstate__(self, state):
        """Set state from pickling."""
        for key, value in state.items():
            setattr(self, key, value)
        # lazy imports
        try:
            from julia import (
                Main,  # noqa: F401
                Pkg,
            )

            Pkg.activate(".")
        except ImportError:
            raise ImportError(
                "Install PyJulia, e.g. via `pip install pypesto[julia]`, "
                "and see the class documentation",
            ) from None
        # Include module if not already included
        _read_source(self.module, self.source_file)

        petab_jl_problem = self.get(self._petab_problem_name)
        self.petab_jl_problem = petab_jl_problem

        # get functions
        fun = self.petab_jl_problem.nllh
        grad = self.petab_jl_problem.grad
        hess = self.petab_jl_problem.hess
        x_names = np.asarray(self.petab_jl_problem.xnames)

        # call the super super constructor
        super(JuliaObjective, self).__init__(fun, grad, hess, x_names)

    def __deepcopy__(self, memodict=None):
        """Deepcopy."""
        return PEtabJlObjective(
            module=self.module,
            source_file=self.source_file,
            petab_problem_name=self._petab_problem_name,
            precompile=False,
        )

    def precompile_model(self, force_compile: bool = False):
        """
        Use Julias PrecompilationTools to precompile the relevant code.

        Only needs to be done once, and speeds up Julia loading drastically.
        """
        directory = os.path.dirname(self.source_file)
        # check whether precompilation is necessary, if the directory exists
        if (
            os.path.exists(f"{directory}/{self.module}_pre")
            and not force_compile
        ):
            logger.info("Precompilation module already exists.")
            return None
        # lazy imports
        try:
            from julia import Main  # noqa: F401
        except ImportError:
            raise ImportError(
                "Install PyJulia, e.g. via `pip install pypesto[julia]`, "
                "and see the class documentation",
            ) from None
        # setting up a local project, where the precompilation will be done in
        from julia import Pkg

        Pkg.activate(".")
        # create a Project f"{self.module}_pre".
        try:
            Pkg.generate(f"{directory}/{self.module}_pre")
        except Exception:
            logger.info("Module is already generated. Skipping generate...")
        # Adjust the precompilation file
        write_precompilation_module(
            module=self.module,
            source_file_orig=self.source_file,
        )
        # add a new line at the top of the original module to use the
        # precompiled module
        with open(self.source_file) as read_f:
            if read_f.readline().endswith("_pre\n"):
                with open("dummy_temp_file.jl", "w+") as write_f:
                    write_f.write(f"using {self.module}_pre\n\n")
                    write_f.write(read_f.read())
                os.remove(self.source_file)
                os.rename("dummy_temp_file.jl", self.source_file)

        try:
            Pkg.develop(path=f"{directory}/{self.module}_pre")
        except Exception:
            logger.info("Module is already developed. Skipping develop...")
        Pkg.activate(f"{directory}/{self.module}_pre/")
        # add dependencies
        Pkg.add("PrecompileTools")
        Pkg.add("OrdinaryDiffEq")
        Pkg.add("PEtab")
        Pkg.add("Sundials")
        Pkg.precompile()


def write_precompilation_module(module, source_file_orig):
    """Write the precompilation module for the PEtabJl module."""
    # read the original source file
    with open(source_file_orig) as f:
        lines = np.array(f.readlines())
    # path to the yaml file
    yaml_path = "\t".join(lines[["yaml" in line for line in lines]])
    # packages
    packages = "\t\t".join(
        lines[[line.startswith("using ") for line in lines]]
    )
    # get everything in between the packages and the end line
    start = int(np.argwhere([line.startswith("using ") for line in lines])[-1])
    end = int(np.argwhere([line.startswith("end") for line in lines])[0])
    petab_loading = "\t\t".join(lines[start:end])

    content = (
        f"module {module}_pre\n\n"
        f"using PrecompileTools\n\n"
        f"# Reduce time for reading a PEtabModel and for "
        f"building a PEtabODEProblem\n"
        f"@setup_workload begin\n"
        f"\t{yaml_path}"
        f"\t@compile_workload begin\n"
        f"\t\t{packages}"
        f"\t\t{petab_loading}"
        f"\tend\n"
        f"end\n\n"
        f"end\n"
    )
    # get the directory of the source file
    directory = os.path.dirname(source_file_orig)
    # write file
    with open(f"{directory}/{module}_pre/src/{module}_pre.jl", "w") as f:
        f.write(content)



"""Enhanced Scatter Search."""

from .ess import ESSExitFlag, ESSOptimizer
from .function_evaluator import (
    FunctionEvaluator,
    FunctionEvaluatorMP,
    FunctionEvaluatorMT,
)
from .refset import RefSet
from .sacess import (
    SacessCmaFactory,
    SacessFidesFactory,
    SacessIpoptFactory,
    SacessOptimizer,
    SacessOptions,
    get_default_ess_options,
)



import logging
from collections.abc import Callable
from typing import Any, Literal

import pypesto.optimize

from ..engine import Task
from ..problem import Problem
from ..result import ProfilerResult
from .options import ProfileOptions
from .walk_along_profile import walk_along_profile

logger = logging.getLogger(__name__)


class ProfilerTask(Task):
    """A parameter likelihood profiling task."""

    def __init__(
        self,
        current_profile: ProfilerResult,
        problem: Problem,
        options: ProfileOptions,
        i_par: int,
        global_opt: float,
        optimizer: "pypesto.optimize.Optimizer",
        create_next_guess: Callable,
        par_direction: Literal[-1, 1],
    ):
        """
        Create the task object.

        Parameters
        ----------
        current_profile:
            The profile which should be computed
        problem:
            The problem to be solved.
        optimizer:
            The optimizer to be used along each profile.
        global_opt:
            log-posterior value of the global optimum
        options:
            Various options applied to the profile optimization.
        create_next_guess:
            Handle of the method which creates the next profile point proposal
        i_par:
            index for the current parameter
        par_direction:
            The direction in which to perform the operation.
            Must be either -1 (descending) or 1 (ascending).
        """
        super().__init__()

        self.optimizer = optimizer
        self.problem = problem
        self.current_profile = current_profile
        self.global_opt = global_opt
        self.create_next_guess = create_next_guess
        self.i_par = i_par
        self.options = options
        self.par_direction = par_direction

    def execute(self) -> dict[str, Any]:
        """Compute profile in descending and ascending direction."""
        logger.debug(
            f"Executing task {'descending' if self.par_direction == -1 else 'ascending'} {self.i_par}."
        )

        # flip profile
        self.current_profile.flip_profile()

        # compute the current profile
        self.current_profile = walk_along_profile(
            current_profile=self.current_profile,
            problem=self.problem,
            par_direction=self.par_direction,
            optimizer=self.optimizer,
            options=self.options,
            create_next_guess=self.create_next_guess,
            global_opt=self.global_opt,
            i_par=self.i_par,
        )

        # return the ProfilerResult and the index of the parameter profiled
        return {
            "profile": self.current_profile,
            "index": self.i_par,
            "par_direction": self.par_direction,
        }



"""Definition of an optimal scaling parameter class."""

from __future__ import annotations

import numpy as np
import pandas as pd

from ...C import (
    C_MATRIX,
    CAT_LB,
    CAT_UB,
    CENSORED,
    CENSORING_BOUNDS,
    CENSORING_TYPES,
    INNER_PARAMETER_BOUNDS,
    INTERVAL_CENSORED,
    LB_INDICES,
    LEFT_CENSORED,
    LIN,
    MEASUREMENT_CATEGORY,
    MEASUREMENT_TYPE,
    NUM_CATEGORIES,
    NUM_CONSTR_FULL,
    NUM_DATAPOINTS,
    NUM_INNER_PARAMS,
    ORDINAL,
    QUANTITATIVE_DATA,
    QUANTITATIVE_IXS,
    REDUCED,
    RIGHT_CENSORED,
    STANDARD,
    SURROGATE_DATA,
    TIME,
    UB_INDICES,
    W_DOT_MATRIX,
    W_MATRIX,
    InnerParameterType,
)
from ..base_problem import (
    AmiciInnerProblem,
    _get_timepoints_with_replicates,
    ix_matrices_from_arrays,
)
from .parameter import OrdinalParameter

try:
    import amici.sim.sundials as asd
    import petab.v1 as petab
    from petab.v1.C import OBSERVABLE_ID, PARAMETER_SEPARATOR
except ImportError:
    pass


class OrdinalProblem(AmiciInnerProblem):
    r"""Inner optimization problem for ordinal or censored data.

    The ordinal inner problem contains the following parameters: surrogate data,
    lower bounds, and upper bounds. All parameters are optimized to minimize the
    distance between the surrogate data and the simulated data while satisfying
    the ordering constraints of the problem. Depending on the method, the problem
    is re-formulated to reduce the number of parameters to estimate.

    Attributes
    ----------
    xs:
        Mapping of (inner) parameter ID to ``InnerParameters``.
    data:
        Measurement data. One matrix (``num_timepoints`` x ``num_observables``)
        per simulation condition. Missing observations as NaN.
    edatas:
        AMICI ``ExpData``\s for each simulation condition.
    groups:
        A dictionary of the groups of the subproblem.
    method:
        A string representing the method of the Optimal Scaling approach, either ``reduced`` or ``standard``.
    """

    def __init__(
        self,
        method: str,
        **kwargs,
    ):
        """Construct."""
        super().__init__(**kwargs)
        self.groups = {}
        self.method = method

        self._initialize_groups()

    def _initialize_groups(self) -> None:
        """Initialize the groups of the subproblem."""
        self.groups = {}

        for group in self.get_groups_for_xs(InnerParameterType.ORDINAL):
            self.groups[group] = {}
            xs = self.get_xs_for_group(group)
            self.groups[group][NUM_CATEGORIES] = len({x.category for x in xs})
            self.groups[group][NUM_DATAPOINTS] = sum(
                [
                    sum([np.sum(ixs) for ixs in x.ixs])
                    for x in self.get_cat_ub_parameters_for_group(group)
                ]
            )

            self.groups[group][SURROGATE_DATA] = np.zeros(
                self.groups[group][NUM_DATAPOINTS]
            )

            self.groups[group][NUM_INNER_PARAMS] = (
                self.groups[group][NUM_DATAPOINTS]
                + 2 * self.groups[group][NUM_CATEGORIES]
            )

            self.groups[group][LB_INDICES] = list(
                range(
                    self.groups[group][NUM_DATAPOINTS],
                    self.groups[group][NUM_DATAPOINTS]
                    + self.groups[group][NUM_CATEGORIES],
                )
            )

            self.groups[group][UB_INDICES] = list(
                range(
                    self.groups[group][NUM_DATAPOINTS]
                    + self.groups[group][NUM_CATEGORIES],
                    self.groups[group][NUM_INNER_PARAMS],
                )
            )

            if all(x.censoring_type is not None for x in xs):
                self.groups[group][MEASUREMENT_TYPE] = CENSORED
                self.groups[group][QUANTITATIVE_IXS] = (
                    self.get_censored_group_quantitative_ixs(xs)
                )
                self.groups[group][QUANTITATIVE_DATA] = np.concatenate(
                    [
                        data_i[mask_i]
                        for data_i, mask_i in zip(
                            self.data,
                            self.groups[group][QUANTITATIVE_IXS],
                            strict=True,
                        )
                    ]
                )
            elif all(x.censoring_type is None for x in xs):
                self.groups[group][MEASUREMENT_TYPE] = ORDINAL

                self.groups[group][NUM_CONSTR_FULL] = (
                    2 * self.groups[group][NUM_DATAPOINTS]
                    + 2 * self.groups[group][NUM_CATEGORIES]
                )

                self.groups[group][C_MATRIX] = self.initialize_c(group)

                self.groups[group][W_MATRIX] = self.initialize_w(group)

                self.groups[group][W_DOT_MATRIX] = self.initialize_w(group)
            else:
                raise ValueError(
                    "Censoring types of optimal scaling parameters of a group "
                    "have to either be all None, or all not None."
                )

    def initialize(self) -> None:
        """Initialize the subproblem."""
        # Initialize all parameter values.
        for x in self.xs.values():
            x.initialize()

        # Initialize the groups.
        for group in self.get_groups_for_xs(InnerParameterType.SPLINE):
            self.groups[group][SURROGATE_DATA] = np.zeros(
                self.groups[group][NUM_DATAPOINTS]
            )

    @staticmethod
    def from_petab_amici(
        petab_problem: petab.Problem,
        amici_model: asd.Model,
        edatas: list[asd.ExpData],
        method: str = None,
    ) -> OrdinalProblem:
        """Construct the inner problem from the `petab_problem`."""
        if not method:
            method = REDUCED

        return optimal_scaling_inner_problem_from_petab_problem(
            petab_problem, amici_model, edatas, method
        )

    def get_interpretable_x_ids(self) -> list[str]:
        """Get IDs of interpretable inner parameters.

        There are no interpretable inner parameters for the ordinal problem.
        """
        return []

    def get_interpretable_x_scales(self) -> list[str]:
        """Get scales of interpretable inner parameters.

        There are no interpretable inner parameters for the ordinal problem.
        """
        return []

    def get_groups_for_xs(self, inner_parameter_type: str) -> list[int]:
        """Get unique list of ``OptimalScalingParameter.group`` values."""
        groups = [x.group for x in self.get_xs_for_type(inner_parameter_type)]
        return list(set(groups))

    def get_xs_for_group(self, group: int) -> list[OrdinalParameter]:
        r"""Get ``OptimalScalingParameter``\s that belong to the given group."""
        return [x for x in self.xs.values() if x.group == group]

    def get_free_xs_for_group(self, group: int) -> list[OrdinalParameter]:
        r"""Get ``OptimalScalingParameter``\s that are free and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.estimate is True
        ]

    def get_fixed_xs_for_group(self, group: int) -> list[OrdinalParameter]:
        r"""Get ``OptimalScalingParameter``\s that are fixed and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.estimate is False
        ]

    def get_cat_ub_parameters_for_group(
        self, group: int
    ) -> list[OrdinalParameter]:
        r"""Get ``OptimalScalingParameter``\s that are category upper boundaries and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.inner_parameter_id[:6] == CAT_UB
        ]

    def get_cat_lb_parameters_for_group(
        self, group: int
    ) -> list[OrdinalParameter]:
        r"""Get ``OptimalScalingParameter``\s that are category lower boundaries and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group and x.inner_parameter_id[:6] == CAT_LB
        ]

    def initialize_c(self, group: int) -> np.ndarray:
        """Initialize the constraints matrix for the group.

        The structure of the constraints matrix is the following: Each row C_i of the matrix C
        represents one optimization constraint as C_i * xi + d(theta, sim) <= 0, where xi is the
        vector of inner paramters (surrogate data, lower bounds, upper bounds)^T, and d is the
        vector of minimal category interval ranges and gaps.

        First `self.groups[group][NUM_DATAPOINTS]` rows constrain the surrogate data to stay
        larger than lower category bounds.
        Then another `self.groups[group][NUM_DATAPOINTS]` rows constrain the surrogate data to
        stay smaller than upper category bounds.
        Then `self.groups[group][NUM_CATEGORIES]` rows constrain the ordering of the categories.
        And lastly, the remaining `self.groups[group][NUM_CATEGORIES]` constrain the lower
        bound to be smaller than the upper bound for each category.
        """
        constr = np.zeros(
            [
                self.groups[group][NUM_CONSTR_FULL],
                self.groups[group][NUM_INNER_PARAMS],
            ]
        )
        data_idx = 0

        # Iterate over categories.
        for cat_idx, category in enumerate(
            self.get_cat_ub_parameters_for_group(group)
        ):
            num_data_in_cat = int(
                np.sum(
                    [
                        np.sum(category.ixs[idx])
                        for idx in range(len(category.ixs))
                    ]
                )
            )

            # Constrain the surrogate data of this category to stay within it.
            for _ in range(num_data_in_cat):
                # lb - y_surr <= 0
                constr[data_idx, data_idx] = -1
                constr[
                    data_idx, cat_idx + self.groups[group][NUM_DATAPOINTS]
                ] = 1

                # y_surr - ub <= 0
                constr[
                    data_idx + self.groups[group][NUM_DATAPOINTS], data_idx
                ] = 1
                constr[
                    data_idx + self.groups[group][NUM_DATAPOINTS],
                    cat_idx
                    + self.groups[group][NUM_DATAPOINTS]
                    + self.groups[group][NUM_CATEGORIES],
                ] = -1
                data_idx += 1

            # Constrain the ordering wrt. neighbouring categories, i.e. ub_i - lb_{i+1} <= 0.
            if cat_idx == 0:
                constr[
                    2 * self.groups[group][NUM_DATAPOINTS] + cat_idx,
                    self.groups[group][NUM_DATAPOINTS] + cat_idx,
                ] = -1
            else:
                constr[
                    2 * self.groups[group][NUM_DATAPOINTS] + cat_idx,
                    self.groups[group][NUM_DATAPOINTS] + cat_idx,
                ] = -1
                constr[
                    2 * self.groups[group][NUM_DATAPOINTS] + cat_idx,
                    self.groups[group][NUM_DATAPOINTS]
                    + self.groups[group][NUM_CATEGORIES]
                    + cat_idx
                    - 1,
                ] = 1

            # Constrain upper bound to be larger than lower bound, i.e. lb_i - ub_i <= 0.
            constr[
                2 * self.groups[group][NUM_DATAPOINTS]
                + self.groups[group][NUM_CATEGORIES]  # - 1
                + cat_idx,
                self.groups[group][LB_INDICES][cat_idx],
            ] = 1
            constr[
                2 * self.groups[group][NUM_DATAPOINTS]
                + self.groups[group][NUM_CATEGORIES]  # - 1
                + cat_idx,
                self.groups[group][UB_INDICES][cat_idx],
            ] = -1

        return constr

    def initialize_w(self, group: int) -> np.ndarray:
        """Initialize the weight matrix for the group."""
        weights = np.diag(
            np.block(
                [
                    np.ones(self.groups[group][NUM_DATAPOINTS]),
                    np.zeros(2 * self.groups[group][NUM_CATEGORIES]),
                ]
            )
        )
        return weights

    def get_w(self, group: int, y_sim_all: np.ndarray) -> np.ndarray:
        """Return the weight matrix of the group."""
        weights = np.diag(
            np.block(
                [
                    np.ones(self.groups[group][NUM_DATAPOINTS])
                    / (np.sum(np.abs(y_sim_all)) + 1e-8),
                    np.zeros(2 * self.groups[group][NUM_CATEGORIES]),
                ]
            )
        )
        return weights

    def get_wdot(
        self, group: int, y_sim_all: np.ndarray, sy_all: np.ndarray
    ) -> np.ndarray:
        """Return the derivative of the weight matrix of a group with respect to an outer parameter."""
        w_dot = np.diag(
            np.block(
                [
                    np.ones(self.groups[group][NUM_DATAPOINTS])
                    * (
                        -1
                        * np.sum(sy_all)
                        / ((np.sum(np.abs(y_sim_all)) + 1e-8) ** 2)
                    ),
                    np.zeros(2 * self.groups[group][NUM_CATEGORIES]),
                ]
            )
        )
        return w_dot

    def get_d(
        self,
        group,
        xs: list[OrdinalParameter],
        y_sim_all: np.ndarray,
        eps: float,
    ) -> np.ndarray:
        """Return vector of minimal gaps and ranges."""
        max_simulation = np.nanmax(y_sim_all)

        interval_range = max_simulation / (2 * len(xs) + 1)
        interval_gap = max_simulation / (4 * (len(xs) - 1) + 1)

        d = np.zeros(self.groups[group][NUM_CONSTR_FULL])

        d[
            2 * self.groups[group][NUM_DATAPOINTS] + 1 : 2
            * self.groups[group][NUM_DATAPOINTS]
            + self.groups[group][NUM_CATEGORIES]
        ] = interval_gap + eps

        d[
            2 * self.groups[group][NUM_DATAPOINTS]
            + self.groups[group][NUM_CATEGORIES] :
        ] = interval_range
        return d

    def get_dd_dtheta(
        self,
        group: int,
        xs: list[OrdinalParameter],
        y_sim_all: np.ndarray,
        sy_all: np.ndarray,
    ) -> np.ndarray:
        """Return the derivative of vector of minimal gaps and ranges with respect to an outer parameter."""
        max_sim_idx = np.argmax(y_sim_all)
        max_sy = sy_all[max_sim_idx]
        dd_dtheta = np.zeros(self.groups[group][NUM_CONSTR_FULL])

        dinterval_range_dtheta = max_sy / (2 * len(xs) + 1)
        dinterval_gap_dtheta = max_sy / (4 * (len(xs) - 1) + 1)

        dd_dtheta[
            2 * self.groups[group][NUM_DATAPOINTS] + 1 : 2
            * self.groups[group][NUM_DATAPOINTS]
            + self.groups[group][NUM_CATEGORIES]
        ] = dinterval_gap_dtheta

        dd_dtheta[
            2 * self.groups[group][NUM_DATAPOINTS]
            + self.groups[group][NUM_CATEGORIES] :
        ] = dinterval_range_dtheta

        return dd_dtheta

    def get_censored_group_quantitative_ixs(
        self, xs: list[OrdinalParameter]
    ) -> list[np.ndarray]:
        r"""Return a list of boolean masks indicating which data points are quantitative.

        For a given group with censored data, return a list of boolean masks indicating
        which data points are not censored, and therefore quantitative.

        Parameters
        ----------
        xs:
            List of ``OptimalScalingParameter``\s of a group with censored data.

        Returns
        -------
        quantitative_ixs:
            List of boolean masks indicating which data points are quantitative.
        """
        # Initialize boolean masks with False and find corresponding observable index.
        quantitative_ixs = [np.full(ixs_i.shape, False) for ixs_i in xs[0].ixs]
        observable_index = xs[0].group - 1

        # Set to True all datapoints of the corresponding observable.
        if np.ndim(quantitative_ixs) == 2:
            quantitative_ixs = [
                np.full(ixs_i.shape, True) for ixs_i in xs[0].ixs
            ]
        else:
            for quantitative_ixs_i in quantitative_ixs:
                quantitative_ixs_i[:, observable_index] = True

        # Set to False for all censored datapoints.
        for x in xs:
            for ixs_i, quantitative_ixs_i in zip(
                x.ixs, quantitative_ixs, strict=True
            ):
                quantitative_ixs_i[ixs_i] = False

        return quantitative_ixs

    def get_inner_parameter_dictionary(self) -> dict:
        """Return a dictionary with inner parameter ids and their values."""
        inner_par_dict = {}
        for x_id, x in self.xs.items():
            inner_par_dict[x_id] = x.value
        return inner_par_dict


def optimal_scaling_inner_problem_from_petab_problem(
    petab_problem: petab.Problem,
    amici_model: asd.Model,
    edatas: list[asd.ExpData],
    method: str,
):
    """Construct the inner problem from the `petab_problem`."""
    # inner parameters
    inner_parameters = optimal_scaling_inner_parameters_from_measurement_df(
        petab_problem.measurement_df, method, amici_model
    )

    # used indices for all measurement specific parameters
    ixs = optimal_scaling_ixs_for_measurement_specific_parameters(
        petab_problem, amici_model, inner_parameters
    )

    # transform experimental data
    data = [asd.ExpDataView(edata)["measurements"] for edata in edatas]

    # matrixify
    ix_matrices = ix_matrices_from_arrays(ixs, data)

    # assign matrices
    for par in inner_parameters:
        par.ixs = ix_matrices[par.inner_parameter_id]

    return OrdinalProblem(
        xs=inner_parameters,
        data=data,
        edatas=edatas,
        method=method,
    )


def optimal_scaling_inner_parameters_from_measurement_df(
    df: pd.DataFrame,
    method: str,
    amici_model: asd.Model,
) -> list[OrdinalParameter]:
    """Create list of inner free parameters from PEtab measurement table dependent on the method provided."""
    df = df.reset_index()

    observable_ids = amici_model.get_observable_ids()

    estimate = get_estimate_for_method(method)
    par_types = [CAT_LB, CAT_UB]

    inner_parameters = []
    lb, ub = INNER_PARAMETER_BOUNDS[InnerParameterType.ORDINAL].values()

    for observable_idx, observable_id in enumerate(observable_ids):
        group = observable_idx + 1

        observable_df = df[df[OBSERVABLE_ID] == observable_id]

        if all(observable_df[MEASUREMENT_TYPE] == ORDINAL):
            # Add optimal scaling parameters for ordinal measurements.
            for par_type, par_estimate in zip(
                par_types, estimate, strict=True
            ):
                for _, row in observable_df.iterrows():
                    par_id = f"{par_type}_{observable_id}_{row[MEASUREMENT_TYPE]}_{int(row[MEASUREMENT_CATEGORY])}"

                    # Create only one set of bound parameters per category of a group.
                    if par_id not in [
                        inner_par.inner_parameter_id
                        for inner_par in inner_parameters
                    ]:
                        inner_parameters.append(
                            OrdinalParameter(
                                inner_parameter_id=par_id,
                                inner_parameter_type=InnerParameterType.ORDINAL,
                                scale=LIN,
                                lb=lb,
                                ub=ub,
                                observable_id=observable_id,
                                category=int(row[MEASUREMENT_CATEGORY]),
                                group=group,
                                estimate=par_estimate,
                            )
                        )
        elif any(observable_df[MEASUREMENT_TYPE].isin(CENSORING_TYPES)):
            # Get df with only censored measurements.
            censored_df = observable_df.loc[
                observable_df[MEASUREMENT_TYPE].isin(CENSORING_TYPES)
            ]
            # Check for unique values in the CENSORING_BOUNDS column and
            # order them with resect to the first float value in the string.
            unique_censoring_bounds = sorted(
                censored_df[CENSORING_BOUNDS].unique(),
                key=lambda x: float(str(x).split(PARAMETER_SEPARATOR)[0]),
            )
            for par_type in par_types:
                for _, row in censored_df.iterrows():
                    category = int(
                        unique_censoring_bounds.index(row[CENSORING_BOUNDS])
                        + 1
                    )
                    par_id = f"{par_type}_{observable_id}_{row[MEASUREMENT_TYPE]}_{category}"
                    # Create only one set of bound parameters per category of a group.
                    if par_id not in [
                        inner_par.inner_parameter_id
                        for inner_par in inner_parameters
                    ]:
                        inner_parameters.append(
                            OrdinalParameter(
                                inner_parameter_id=par_id,
                                inner_parameter_type=InnerParameterType.ORDINAL,
                                scale=LIN,
                                lb=lb,
                                ub=ub,
                                observable_id=observable_id,
                                category=category,
                                group=group,
                                estimate=False,
                                censoring_type=row[MEASUREMENT_TYPE],
                            )
                        )
                        _add_value_to_censored_bound_parameter(
                            inner_parameters[-1], row, par_type
                        )

    inner_parameters.sort(key=lambda x: (x.group, x.category))

    return inner_parameters


def get_estimate_for_method(method: str) -> tuple[bool, bool]:
    """Return which inner parameters to estimate dependent on the method provided."""
    estimate_ub = True
    estimate_lb = False

    if method == STANDARD:
        estimate_lb = True

    return estimate_lb, estimate_ub


def optimal_scaling_ixs_for_measurement_specific_parameters(
    petab_problem: petab.Problem,
    amici_model: asd.Model,
    inner_parameters: list[OrdinalParameter],
) -> dict[str, list[tuple[int, int, int]]]:
    """Create mapping of parameters to measurements.

    Returns
    -------
    A dictionary mapping parameter ID to a list of
    `(condition index, time index, observable index)` tuples in which this
    output parameter is used. For each condition, the time index refers to
    a sorted list of non-unique time points for which there are measurements.
    """
    ixs_for_par = {}
    observable_ids = amici_model.get_observable_ids()

    simulation_conditions = (
        petab_problem.get_simulation_conditions_from_measurement_df()
    )
    measurement_df = petab_problem.measurement_df
    # Get unique censoring bounds for each observable.
    unique_censoring_bounds_per_observable = {}
    if CENSORING_BOUNDS in measurement_df.columns:
        for observable_id in observable_ids:
            observable_df = measurement_df[
                measurement_df[OBSERVABLE_ID] == observable_id
            ]
            censored_observable_df = observable_df.loc[
                observable_df[MEASUREMENT_TYPE].isin(CENSORING_TYPES)
            ]
            unique_censoring_bounds_per_observable[observable_id] = sorted(
                censored_observable_df[CENSORING_BOUNDS].unique(),
                key=lambda x: float(str(x).split(PARAMETER_SEPARATOR)[0]),
            )

    for condition_ix, condition in simulation_conditions.iterrows():
        # measurement table for current condition
        df_for_condition = petab.get_rows_for_condition(
            measurement_df=measurement_df, condition=condition
        )

        # unique sorted list of timepoints
        timepoints = sorted(df_for_condition[TIME].unique().astype(float))
        # non-unique sorted list of timepoints
        timepoints_w_reps = _get_timepoints_with_replicates(
            measurement_df=df_for_condition
        )

        for time in timepoints:
            # subselect measurements for time `time`
            df_for_time = df_for_condition[df_for_condition[TIME] == time]
            time_ix_0 = timepoints_w_reps.index(time)

            # remember used time indices for each observable
            time_ix_for_obs_ix = {}

            # iterate over measurements
            for _, measurement in df_for_time.iterrows():
                # extract observable index
                observable_ix = observable_ids.index(
                    measurement[OBSERVABLE_ID]
                )

                # as the time indices have to account for replicates, we need
                #  to track which time indices have already been assigned for
                #  the current observable
                if observable_ix in time_ix_for_obs_ix:
                    # a replicate
                    time_ix_for_obs_ix[observable_ix] += 1
                else:
                    # the first measurement for this `(observable, timepoint)`
                    time_ix_for_obs_ix[observable_ix] = time_ix_0
                time_w_reps_ix = time_ix_for_obs_ix[observable_ix]

                inner_par_ids_for_meas = get_inner_par_ids_for_measurement(
                    measurement,
                    inner_parameters,
                    unique_censoring_bounds_per_observable,
                )

                # try to insert if hierarchical parameter
                if inner_par_ids_for_meas:
                    for override in inner_par_ids_for_meas:
                        ixs_for_par.setdefault(override, []).append(
                            (condition_ix, time_w_reps_ix, observable_ix)
                        )
    return ixs_for_par


def get_inner_par_ids_for_measurement(
    measurement: dict,
    inner_parameters: list[OrdinalParameter],
    unique_censoring_bounds_per_observable: dict[str, list[float]],
):
    """Return inner parameter ids of parameters which are related to the measurement."""
    if measurement[MEASUREMENT_TYPE] == ORDINAL:
        return [
            inner_par.inner_parameter_id
            for inner_par in inner_parameters
            if inner_par.category == measurement[MEASUREMENT_CATEGORY]
            and inner_par.observable_id == measurement[OBSERVABLE_ID]
        ]
    elif measurement[MEASUREMENT_TYPE] in [
        LEFT_CENSORED,
        INTERVAL_CENSORED,
        RIGHT_CENSORED,
    ]:
        unique_censoring_bounds = unique_censoring_bounds_per_observable[
            measurement[OBSERVABLE_ID]
        ]
        measurement_category = int(
            unique_censoring_bounds.index(measurement[CENSORING_BOUNDS]) + 1
        )
        return [
            inner_par.inner_parameter_id
            for inner_par in inner_parameters
            if inner_par.observable_id == measurement[OBSERVABLE_ID]
            and inner_par.category == measurement_category
        ]


def _add_value_to_censored_bound_parameter(
    inner_parameter: OrdinalParameter,
    row: pd.Series,
    par_type: str,
) -> None:
    if row[MEASUREMENT_TYPE] == LEFT_CENSORED and par_type == CAT_LB:
        inner_parameter.value = 0.0
    elif row[MEASUREMENT_TYPE] == LEFT_CENSORED and par_type == CAT_UB:
        inner_parameter.value = float(row[CENSORING_BOUNDS])

    elif row[MEASUREMENT_TYPE] == RIGHT_CENSORED and par_type == CAT_LB:
        inner_parameter.value = float(row[CENSORING_BOUNDS])
    elif row[MEASUREMENT_TYPE] == RIGHT_CENSORED and par_type == CAT_UB:
        inner_parameter.value = np.inf

    elif row[MEASUREMENT_TYPE] == INTERVAL_CENSORED and par_type == CAT_LB:
        inner_parameter.value = float(
            row[CENSORING_BOUNDS].split(PARAMETER_SEPARATOR)[0]
        )
    elif row[MEASUREMENT_TYPE] == INTERVAL_CENSORED and par_type == CAT_UB:
        inner_parameter.value = float(
            row[CENSORING_BOUNDS].split(PARAMETER_SEPARATOR)[1]
        )



from functools import partial

import numpy as np

try:
    import petab.v1 as petab
    from petab.v1.C import (
        ESTIMATE,
        NOISE_PARAMETERS,
        OBSERVABLE_ID,
    )
except ImportError:
    petab = None

from ..C import (
    CENSORED,
    CENSORING_TYPES,
    MEASUREMENT_TYPE,
    ORDINAL,
    PARAMETER_TYPE,
    RELATIVE,
    SEMIQUANTITATIVE,
    InnerParameterType,
)
from ..problem import Problem
from ..startpoint import CheckedStartpoints


def get_petab_non_quantitative_data_types(
    petab_problem: petab.Problem,
) -> set[str]:
    """
    Get the data types from the PEtab problem.

    Parameters
    ----------
    petab_problem:
        The PEtab problem.

    Returns
    -------
    data_types:
        A list of the data types.
    """
    non_quantitative_data_types = set()
    caught_observables = set()
    # For ordinal, censored and semiquantitative data, search
    # for the corresponding data types in the measurement table
    meas_df = petab_problem.measurement_df
    if MEASUREMENT_TYPE in meas_df.columns:
        petab_data_types = meas_df[MEASUREMENT_TYPE].unique()
        for data_type in [ORDINAL, SEMIQUANTITATIVE] + CENSORING_TYPES:
            if data_type in petab_data_types:
                non_quantitative_data_types.add(
                    CENSORED if data_type in CENSORING_TYPES else data_type
                )
                caught_observables.update(
                    set(
                        meas_df[meas_df[MEASUREMENT_TYPE] == data_type][
                            OBSERVABLE_ID
                        ]
                    )
                )

    # For relative data, search for parameters to estimate with
    # a scaling/offset/sigma parameter type
    if PARAMETER_TYPE in petab_problem.parameter_df.columns:
        # get the df with non-nan parameter types
        par_df = petab_problem.parameter_df[
            petab_problem.parameter_df[PARAMETER_TYPE].notna()
        ]
        for par_id, row in par_df.iterrows():
            if not row[ESTIMATE]:
                continue
            if row[PARAMETER_TYPE] in [
                InnerParameterType.SCALING,
                InnerParameterType.OFFSET,
            ]:
                non_quantitative_data_types.add(RELATIVE)

            # For sigma parameters, we need to check if they belong
            # to an observable with a non-quantitative data type
            elif row[PARAMETER_TYPE] == InnerParameterType.SIGMA:
                corresponding_observables = set(
                    meas_df[meas_df[NOISE_PARAMETERS] == par_id][OBSERVABLE_ID]
                )
                if not (corresponding_observables & caught_observables):
                    non_quantitative_data_types.add(RELATIVE)

    # TODO this can be made much shorter if the relative measurements
    # are also specified in the measurement table, but that would require
    # changing the PEtab format of a lot of benchmark models.

    if len(non_quantitative_data_types) == 0:
        return None
    return non_quantitative_data_types


class PetabStartpoints(CheckedStartpoints):
    """Startpoint method for PEtab problems.

    Samples optimization startpoints from the distributions defined in the
    provided PEtab problem. The PEtab-problem is copied.
    """

    def __init__(self, petab_problem: petab.Problem, **kwargs):
        super().__init__(**kwargs)
        self._parameter_df = petab_problem.parameter_df.copy()
        self._priors: list[tuple] | None = None
        self._free_ids: list[str] | None = None

    def _setup(
        self,
        pypesto_problem: Problem,
    ):
        """Update priors if necessary.

        Check if ``problem.x_free_indices`` changed since last call, and if so,
        get the corresponding priors from PEtab.
        """
        current_free_ids = np.asarray(pypesto_problem.x_names)[
            pypesto_problem.x_free_indices
        ]

        if (
            self._priors is not None
            and len(current_free_ids) == len(self._free_ids)
            and np.all(current_free_ids == self._free_ids)
        ):
            # no need to update
            return

        # update priors
        self._free_ids = current_free_ids
        id_to_prior = dict(
            zip(
                self._parameter_df.index[self._parameter_df[ESTIMATE] == 1],
                petab.parameters.get_priors_from_df(
                    self._parameter_df, mode=petab.INITIALIZATION
                ),
                strict=True,
            )
        )

        self._priors = list(map(id_to_prior.__getitem__, current_free_ids))

    def __call__(
        self,
        n_starts: int,
        problem: Problem,
    ) -> np.ndarray:
        """Call the startpoint method."""
        # Update the list of priors if needed
        self._setup(pypesto_problem=problem)

        return super().__call__(n_starts, problem)

    def sample(
        self,
        n_starts: int,
        lb: np.ndarray,
        ub: np.ndarray,
        priors=None,
    ) -> np.ndarray:
        """Actual startpoint sampling.

        Must only be called through `self.__call__` to ensure that the list of priors
        matches the currently free parameters in the :class:`pypesto.Problem`.
        """
        sampler = partial(petab.sample_from_prior, n_starts=n_starts)
        startpoints = list(map(sampler, self._priors))

        return np.array(startpoints).T



"""ReferenceSet functionality for scatter search."""

from typing import Any

import numpy as np

from .function_evaluator import FunctionEvaluator


class RefSet:
    """Scatter search reference set.

    Attributes
    ----------
    dim:
        Reference set size
    evaluator:
        Function evaluator
    x:
        Parameters in the reference set
    fx:
        Function values at the parameters in the reference set
    n_stuck:
        Counts the number of times a refset member did not lead to an
        improvement in the objective (length: ``dim``).
    """

    def __init__(
        self,
        dim: int,
        evaluator: FunctionEvaluator,
        x: np.ndarray | None = None,
        fx: np.ndarray | None = None,
    ):
        """Construct.

        Parameters
        ----------
        dim:
            Reference set size
        evaluator:
            Function evaluator
        x:
            Initial RefSet parameters.
        fx:
            Function values corresponding to entries in x. Must be provided if
            and only if ``x`` is not ``None``.
        """
        if (x is not None and fx is None) or (x is None and fx is not None):
            raise ValueError(
                "Either both or neither of `x` and `fx` should be provided"
            )

        if dim < 3:
            raise ValueError("RefSet dimension has to be at least 3.")
        self.dim = dim
        self.evaluator = evaluator
        # \epsilon in [PenasGon2017]_
        self.proximity_threshold = 1e-3

        if x is None:
            self.x = self.fx = None
        else:
            self.x = x
            self.fx = fx

        self.n_stuck = np.zeros(shape=[dim])
        self.attributes: dict[Any, np.ndarray] = {}

    def __repr__(self):
        fx = (
            f", fx=[{np.min(self.fx)} ... {np.max(self.fx)}]"
            if self.fx is not None and len(self.fx) >= 2
            else ""
        )
        return f"RefSet(dim={self.dim}{fx})"

    def sort(self):
        """Sort RefSet by quality."""
        order = np.argsort(self.fx)
        self.fx = self.fx[order]
        self.x = self.x[order]
        self.n_stuck = self.n_stuck[order]
        for attribute_name, attribute_values in self.attributes.items():
            self.attributes[attribute_name] = attribute_values[order]

    def initialize_random(
        self,
        n_diverse: int,
    ):
        """Create an initial reference set from random parameters.

        Sample ``n_diverse`` random points, populate half of the RefSet using
        the best solutions and fill the rest with random points.
        """
        # sample n_diverse points
        x_diverse, fx_diverse = self.evaluator.multiple_random(n_diverse)
        self.initialize_from_array(x_diverse=x_diverse, fx_diverse=fx_diverse)

    def initialize_from_array(
        self, x_diverse: np.ndarray, fx_diverse: np.ndarray
    ):
        """Create an initial reference set using the provided points.

        Populate half of the RefSet using the best given solutions and fill the
        rest with a random selection from the remaining points.
        """
        if len(x_diverse) != len(fx_diverse):
            raise ValueError(
                "Lengths of `x_diverse` and `fx_diverse` do not match."
            )
        if self.dim > len(x_diverse):
            raise ValueError(
                "Cannot create RefSet with dimension "
                f"{self.dim} from only {len(x_diverse)} points."
            )

        self.fx = np.full(shape=(self.dim,), fill_value=np.inf)
        self.x = np.full(
            shape=(self.dim, self.evaluator.problem.dim), fill_value=np.nan
        )

        # create initial refset with 50% best values
        num_best = int(self.dim / 2)
        order = np.argsort(fx_diverse)
        self.x[:num_best] = x_diverse[order[:num_best]]
        self.fx[:num_best] = fx_diverse[order[:num_best]]

        # ... and 50% random points
        random_idxs = np.random.choice(
            order[num_best:], size=self.dim - num_best, replace=False
        )
        self.x[num_best:] = x_diverse[random_idxs]
        self.fx[num_best:] = fx_diverse[random_idxs]

    def prune_too_close(self):
        """Prune too similar RefSet members.

        Replace a parameter vector if its maximum relative difference to a
        better parameter vector is below the given threshold.

        Assumes RefSet is sorted.
        """
        # Compare [PenasGon2007]
        #  Note that the main text states that distance between the two points
        #  is normalized to the bounds of the search space. However,
        #  Algorithm 1, line 9 normalizes to x_j instead. The accompanying
        #  code does normalize to max(abs(x_i), abs(x_j)).
        # Normalizing to the bounds of the search space seems more reasonable.
        #  Otherwise, for a parameter with bounds [lb, ub],
        #  where (ub-lb)/ub < proximity_threshold, we would never find an
        #  admissible point.
        x = self.x
        ub, lb = self.evaluator.problem.ub, self.evaluator.problem.lb

        def normalize(x):
            """Normalize parameter vector to the bounds of the search space."""
            return (x - lb) / (ub - lb)

        for i in range(self.dim):
            for j in range(i + 1, self.dim):
                # check proximity
                # zero-division may occur here
                with np.errstate(divide="ignore", invalid="ignore"):
                    while (
                        np.max(np.abs(normalize(x[i]) - normalize(x[j])))
                        <= self.proximity_threshold
                    ):
                        # too close. replace x_j.
                        x[j], self.fx[j] = self.evaluator.single_random()
                        self.sort()

    def update(self, i: int, x: np.ndarray, fx: float):
        """Update a RefSet entry."""
        self.x[i] = x
        self.fx[i] = fx
        self.n_stuck[i] = 0

    def replace_by_random(self, i: int):
        """Replace the RefSet member with the given index by a random point."""
        self.x[i], self.fx[i] = self.evaluator.single_random()
        self.n_stuck[i] = 0

    def add_attribute(self, name: str, values: np.ndarray):
        """
        Add an attribute array to the refset members.

        An attribute can be any 1D array of the same length as the refset.
        The added array will be sorted together with the refset members.
        """
        if len(values) != self.dim:
            raise ValueError("Attribute length does not match refset length.")
        self.attributes[name] = values

    def resize(self, new_dim: int):
        """
        Resize the refset.

        If the dimension does not change, do nothing.
        If size is decreased, drop entries from the end (i.e., the worst
        values, assuming it is sorted). If size is increased, the new
        entries are filled with randomly sampled parameters and the refset is
        sorted.

        NOTE: Any attributes are just truncated or filled with zeros.
        """
        if new_dim == self.dim:
            return

        if new_dim < self.dim:
            # shrink
            self.fx = self.fx[:new_dim]
            self.x = self.x[:new_dim]
            self.n_stuck = self.n_stuck[:new_dim]
            for attribute_name, attribute_values in self.attributes.items():
                self.attributes[attribute_name] = attribute_values[:new_dim]
            self.dim = new_dim
        else:
            # grow
            new_x, new_fx = self.evaluator.multiple_random(new_dim - self.dim)
            self.fx = np.append(self.fx, new_fx)
            self.x = np.vstack((self.x, new_x))
            self.n_stuck = np.append(
                self.n_stuck, np.zeros(shape=(new_dim - self.dim))
            )
            for attribute_name, attribute_values in self.attributes.items():
                self.attributes[attribute_name] = np.append(
                    attribute_values, np.zeros(shape=(new_dim - self.dim))
                )
            self.dim = new_dim
            self.sort()



"""Prior-based sampling."""

import logging

import numpy as np

from ..objective import NegLogParameterPriors, NegLogPriors
from .base import CheckedStartpoints
from .uniform import uniform

logger = logging.getLogger(__name__)


class PriorStartpoints(CheckedStartpoints):
    """Generate startpoints from prior distribution.

    Samples from the prior distribution if available, otherwise falls back
    to uniform sampling. Ensures all samples are within bounds by resampling
    out-of-bounds values.
    """

    def __init__(
        self,
        use_guesses: bool = True,
        check_fval: bool = False,
        check_grad: bool = False,
    ):
        """Initialize.

        Parameters
        ----------
        use_guesses:
            Whether to use guesses provided in the problem.
        check_fval:
            Whether to check function values at the startpoint, and resample
            if not finite.
        check_grad:
            Whether to check gradients at the startpoint, and resample
            if not finite.
        """
        super().__init__(
            use_guesses=use_guesses,
            check_fval=check_fval,
            check_grad=check_grad,
        )

    def sample(
        self,
        n_starts: int,
        lb: np.ndarray,
        ub: np.ndarray,
        priors: NegLogParameterPriors | NegLogPriors | None = None,
        max_iterations: int = 100,
    ) -> np.ndarray:
        """Sample startpoints from prior or uniform distribution.

        Parameters
        ----------
        n_starts: Number of startpoints to generate.
        lb: Lower parameter bound.
        ub: Upper parameter bound.
        priors: Parameter priors. If available, samples from priors;
            otherwise falls back to uniform sampling.
        max_iterations: Maximum number of resampling iterations to ensure
            all samples are within bounds.

        Returns
        -------
        xs: Startpoints, shape (n_starts, n_par).
        """
        # Fallback to uniform sampling if no priors available
        if priors is None:
            return uniform(n_starts=n_starts, lb=lb, ub=ub)

        n_par = len(lb)
        lb_reshaped = lb.reshape(1, -1)
        ub_reshaped = ub.reshape(1, -1)

        # Initialize samples and mark all rows for initial sampling
        samples = np.zeros((n_starts, n_par))
        rows_to_resample = np.ones(n_starts, dtype=bool)

        iteration = 0
        while True:
            n_resample = rows_to_resample.sum()
            if n_resample == 0:
                break

            # Generate new samples for the selected rows
            new_samples = np.zeros((n_resample, n_par))

            # Sample from priors where available
            prior_samples_dict = priors.sample(n_samples=n_resample)
            for param_index in prior_samples_dict.keys():
                new_samples[:, param_index] = prior_samples_dict[param_index]

            # Fill in uniform samples for parameters without priors
            for param_index in range(n_par):
                if param_index not in prior_samples_dict.keys():
                    new_samples[:, param_index] = np.random.uniform(
                        lb[param_index], ub[param_index], n_resample
                    )

            # Replace selected rows in the main array
            samples[rows_to_resample] = new_samples

            # Check bounds
            out_of_bounds = np.logical_or(
                samples < lb_reshaped, samples > ub_reshaped
            )

            # If all samples are within bounds, we are done
            if not np.any(out_of_bounds):
                break

            iteration += 1
            if iteration >= max_iterations:
                samples = np.clip(samples, lb_reshaped, ub_reshaped)
                logger.warning(
                    "Maximum startpoint resampling iterations reached. "
                    "Some samples were clipped to bounds."
                )
                break

            # Resample only rows that are still out-of-bounds
            rows_to_resample = np.any(out_of_bounds, axis=1)

        return samples



"""Objective function for RoadRunner models.

Currently does not support sensitivities.
"""

from __future__ import annotations

from collections import OrderedDict
from collections.abc import Sequence

import numpy as np

from ...C import MODE_FUN, MODE_RES, ROADRUNNER_INSTANCE, X_NAMES, ModeType
from ..base import ObjectiveBase
from .roadrunner_calculator import RoadRunnerCalculator
from .utils import ExpData, SolverOptions

try:
    from petab.v1 import Problem as PetabProblem
    from petab.v1.parameter_mapping import ParMappingDictQuadruple
except ImportError:
    petab = None
try:
    import roadrunner
except ImportError:
    roadrunner = None


class RoadRunnerObjective(ObjectiveBase):
    """Objective function for RoadRunner models.

    Currently does not support sensitivities.
    """

    def __init__(
        self,
        rr: roadrunner.RoadRunner,
        edatas: Sequence[ExpData] | ExpData,
        parameter_mapping: list[ParMappingDictQuadruple],
        petab_problem: PetabProblem,
        calculator: RoadRunnerCalculator | None = None,
        x_ids: Sequence[str] | None = None,
        x_names: Sequence[str] | None = None,
        solver_options: SolverOptions | None = None,
    ):
        """Initialize the RoadRunner objective function.

        Parameters
        ----------
        rr:
            RoadRunner instance for simulation.
        edatas:
            The experimental data. If a list is passed, its entries correspond
            to multiple experimental conditions.
        parameter_mapping:
            Mapping of optimization parameters to model parameters. Format as
            created by `petab.get_optimization_to_simulation_parameter_mapping`.
            The default is just to assume that optimization and simulation
            parameters coincide.
        petab_problem:
            The corresponding PEtab problem. Needed to calculate NLLH.
            Might be removed later.
        calculator:
            The calculator to use. If None, a new instance is created.
        x_ids:
            IDs of Roadrunner parameters. Includes fixed parameters as well.
        x_names:
            Names of optimization parameters.
        """
        self.roadrunner_instance = rr
        # make sure edatas are a list
        if isinstance(edatas, ExpData):
            edatas = [edatas]
        self.edatas = edatas
        self.parameter_mapping = parameter_mapping
        self.petab_problem = petab_problem
        if calculator is None:
            calculator = RoadRunnerCalculator()
        self.calculator = calculator
        if solver_options is None:
            solver_options = SolverOptions()
        self.solver_options = solver_options
        if x_ids is None:
            x_ids = list(rr.model.getGlobalParameterIds())
        self.x_ids = x_ids
        if x_names is None:
            x_names = x_ids
        super().__init__(x_names=x_names)

    def get_config(self) -> dict:
        """Return basic information of the objective configuration."""
        info = super().get_config()
        info["solver_options"] = repr(self.solver_options)
        info[X_NAMES] = self.x_names
        info[ROADRUNNER_INSTANCE] = self.roadrunner_instance.getInfo()
        return info

    # TODO: Check whether we need some sort of pickling

    def __call__(
        self,
        x: np.ndarray,
        sensi_orders: tuple[int, ...] = (0,),
        mode: ModeType = MODE_FUN,
        return_dict: bool = False,
        **kwargs,
    ) -> float | np.ndarray | dict:
        """See :class:`ObjectiveBase` documentation."""
        return super().__call__(x, sensi_orders, mode, return_dict, **kwargs)

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: tuple[int, ...],
        mode: ModeType,
        return_dict: bool,
        edatas: Sequence[ExpData] | None = None,
        parameter_mapping: list[ParMappingDictQuadruple] | None = None,
    ) -> dict:
        """
        Call objective function without pre- or post-processing and formatting.

        Returns
        -------
        result:
            A dict containing the results.
        """
        # fill in values if not passed
        if edatas is None:
            edatas = self.edatas
        if parameter_mapping is None:
            parameter_mapping = self.parameter_mapping
        # convert x to dictionary
        x = OrderedDict(zip(self.x_ids, x, strict=True))
        ret = self.calculator(
            x_dct=x,
            mode=mode,
            roadrunner_instance=self.roadrunner_instance,
            edatas=edatas,
            x_ids=self.x_names,
            parameter_mapping=parameter_mapping,
            petab_problem=self.petab_problem,
            solver_options=self.solver_options,
        )
        return ret

    def check_sensi_orders(
        self,
        sensi_orders: tuple[int, ...],
        mode: ModeType,
    ) -> bool:
        """See :class:`ObjectiveBase` documentation."""
        if not sensi_orders:
            return True
        sensi_order = max(sensi_orders)
        max_sensi_order = 0

        return sensi_order <= max_sensi_order

    def check_mode(self, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        return mode in [MODE_FUN, MODE_RES]



"""Contains the PetabJlImporter class."""

from __future__ import annotations

import logging
import os.path
from collections.abc import Iterable

import numpy as np

from pypesto.objective.julia import PEtabJlObjective
from pypesto.problem import Problem

logger = logging.getLogger(__name__)

PEtabProblemJl = ["PEtab.jl::PEtabODEProblem"]


class PetabJlImporter:
    """
    Importer for PEtab models in Julia, using PEtab.jl.

    Create an `objective.JuliaObjective` or a `pypesto.Problem` from PEtab
    files or from a julia module.
    """

    def __init__(
        self,
        module: str = None,
        source_file: str = None,
        petab_problem_name: str = "petabProblem",
    ):
        """
        Initialize importer.

        Parameters
        ----------
        module:
            Name of the Julia model
        source_file:
            Path to the Julia source file.
        petab_problem:
            Wrapper around the PEtab.jl problem.
        """
        self.module = module
        self.source_file = source_file
        self._petab_problem_name = petab_problem_name
        # placeholder for the petab.jl problem
        self.petab_jl_problem = None

    @staticmethod
    def from_yaml(
        yaml_file: str,
        ode_solver_options: dict | None = None,
        gradient_method: str | None = None,
        hessian_method: str | None = None,
        sparse_jacobian: bool | None = None,
        verbose: bool | None = None,
        directory: str | None = None,
    ) -> PetabJlImporter:
        """
        Create a `PetabJlImporter` from a yaml file.

        Writes the Julia module to a file in `directory` and returns a
        `PetabJlImporter` for that module.

        Parameters
        ----------
        yaml_file:
            The yaml file of the PEtab problem
        ode_solver_options:
            Dictionary like options for the ode solver in julia
        gradient_method, hessian_method:
            Julia methods to compute gradient and hessian
        sparse_jacobian:
            Whether to compute sparse Jacobians
        verbose:
            Whether to have a more informative log.
        directory:
            Where to write the julia file, defaults to the directory of the
            yaml file.
        """
        # get default values
        options = _get_default_options(
            ode_solver_options=ode_solver_options,
            gradient_method=gradient_method,
            hessian_method=hessian_method,
            sparse_jacobian=sparse_jacobian,
            verbose=verbose,
        )

        # write julia module
        source_file, module = _write_julia_file(
            yaml_file=yaml_file, options=options, directory=directory
        )

        return PetabJlImporter(
            module=module,
            source_file=source_file,
        )

    def create_objective(
        self, precompile: bool | None = True
    ) -> PEtabJlObjective:
        """
        Create a `pypesto.objective.PEtabJlObjective` from the PEtab.jl problem.

        The objective function will be the negative log likelihood or the
        negative log posterior, depending on the PEtab.jl problem.

        Parameters
        ----------
        precompile:
            Whether to precompile the julia module for speed up in
            multistart optimization.
        """
        # lazy imports
        try:
            from julia import Main  # noqa: F401
        except ImportError:
            raise ImportError(
                "Install PyJulia, e.g. via `pip install pypesto[julia]`, "
                "and see the class documentation",
            ) from None
        if self.source_file is None:
            self.source_file = f"{self.module}.jl"

        if not os.path.exists(self.source_file):
            raise ValueError(
                "The julia file does not exist. You can create "
                "it from a petab yaml file path using "
                "`PetabJlImporter.from_yaml(yaml_file)`"
            )

        obj = PEtabJlObjective(
            module=self.module,
            source_file=self.source_file,
            petab_problem_name=self._petab_problem_name,
            precompile=precompile,
        )

        self.petab_jl_problem = obj.petab_jl_problem
        return obj

    def create_problem(
        self,
        x_guesses: Iterable[float] | None = None,
        lb_init: np.ndarray | list[float] | None = None,
        ub_init: np.ndarray | list[float] | None = None,
        precompile: bool | None = True,
    ) -> Problem:
        """
        Create a `pypesto.Problem` from the PEtab.jl problem.

        Parameters
        ----------
        x_guesses:
            Guesses for the parameter values, shape (g, dim), where g denotes the
            number of guesses. These are used as start points in the optimization.
        lb_init, ub_init:
            The lower and upper bounds for initialization, typically for defining
            search start points.
            If not set, set to lb, ub.
        precompile:
            Whether to precompile the julia module for speed up in
            multistart optimization.
        """
        obj = self.create_objective(precompile=precompile)
        lb = np.asarray(self.petab_jl_problem.lower_bounds)
        ub = np.asarray(self.petab_jl_problem.upper_bounds)

        return Problem(
            objective=obj,
            lb=lb,
            ub=ub,
            x_guesses=x_guesses,
            x_names=obj.x_names,
            lb_init=lb_init,
            ub_init=ub_init,
        )


def _get_default_options(
    ode_solver_options: dict | None = None,
    gradient_method: str | None = None,
    hessian_method: str | None = None,
    sparse_jacobian: str | None = None,
    verbose: str | None = None,
) -> dict:
    """
    If values are not specified, get default values for the options.

    Additionally check that the values are valid.

    Parameters
    ----------
    ode_solver_options:
        Options for the ODE solver.
    gradient_method:
        Method for gradient calculation.
    hessian_method:
        Method for hessian calculation.
    sparse_jacobian:
        Whether the jacobian should be sparse.
    verbose:
        Whether to print verbose output.

    Returns
    -------
    dict:
        The options.
    """
    # get default values
    if ode_solver_options is None:
        ode_solver_options = {
            "solver": "Rodas5P",
            "abstol": 1e-8,
            "reltol": 1e-8,
            "maxiters": "Int64(1e4)",
        }
    if not ode_solver_options["solver"].endswith("()"):
        ode_solver_options["solver"] += "()"  # add parentheses
    if gradient_method is None:
        gradient_method = "nothing"
    if hessian_method is None:
        hessian_method = "nothing"
    if sparse_jacobian is None:
        sparse_jacobian = "nothing"
    if verbose is None:
        verbose = "true"

    # check values for gradient_method and hessian_method
    allowed_gradient_methods = [
        "ForwardDiff",
        "ForwardEquations",
        "Adjoint",
        "Zygote",
    ]
    if gradient_method not in allowed_gradient_methods:
        logger.warning(
            f"gradient_method {gradient_method} is not in "
            f"{allowed_gradient_methods}. Defaulting to ForwardDiff."
        )
        gradient_method = "ForwardDiff"
    allowed_hessian_methods = ["ForwardDiff", "BlocForwardDiff", "GaussNewton"]
    if hessian_method not in allowed_hessian_methods:
        logger.warning(
            f"hessian_method {hessian_method} is not in "
            f"{allowed_hessian_methods}. Defaulting to ForwardDiff."
        )
        hessian_method = "ForwardDiff"

    # fill options
    options = {
        "ode_solver_options": ode_solver_options,
        "gradient_method": gradient_method,
        "hessian_method": hessian_method,
        "sparse_jacobian": sparse_jacobian,
        "verbose": verbose,
    }
    return options


def _write_julia_file(
    yaml_file: str, options: dict, directory: str
) -> tuple[str, str]:
    """
    Write the Julia file.

    Parameters
    ----------
    yaml_file:
        The yaml file of the PEtab problem.
    options:
        The options.
    dir:
        The directory to write the file to.

    Returns
    -------
    source_file:
        The name/path of the file.
    module:
        The module name.
    """
    if directory is None:
        directory = os.path.dirname(yaml_file)  # directory of the yaml file
    source_file = os.path.join(directory, "PEtabJl_module.jl")
    module = "MyPEtabJlModule"

    link_to_options = (
        "https://sebapersson.github.io/"
        "PEtab.jl/dev/API_choosen/#PEtab.setupPEtabODEProblem"
    )
    odeSolvOpt_str = ", ".join(
        [f"{k}={v}" for k, v in options["ode_solver_options"].items()]
    )
    # delete "solver=" from string
    odeSolvOpt_str = odeSolvOpt_str.replace("solver=", "")

    content = (
        f"module {module}\n\n"
        f"using OrdinaryDiffEq\n"
        f"using Sundials\n"
        f"using PEtab\n\n"
        f'pathYaml = "{yaml_file}"\n'
        f"petabModel = PEtabModel(pathYaml, verbose=true)\n\n"
        f"# A full list of options for PEtabODEProblem can be "
        f"found at {link_to_options}\n"
        f"petabProblem = PEtabODEProblem(\n\t"
        f"petabModel,\n\t"
        f"odesolver=ODESolver({odeSolvOpt_str}),\n\t"
        f"gradient_method=:{options['gradient_method']},\n\t"
        f"hessian_method=:{options['hessian_method']},\n\t"
        f"sparse_jacobian={options['sparse_jacobian']},\n\t"
        f"verbose={options['verbose']}\n)\n\nend\n"
    )
    # write file
    with open(source_file, "w") as f:
        f.write(content)

    return source_file, module



"""
Jax models interface.

Adds an interface for the construction of loss functions
incorporating jax models. This permits computation of derivatives using a
combination of objective based methods and jax based autodiff.
"""

import copy
from collections.abc import Callable
from functools import partial

import numpy as np

from ...C import MODE_FUN, ModeType
from ..base import ObjectiveBase, ResultDict

try:
    import jax
    import jax.numpy as jnp
    from jax import custom_jvp
except ImportError:
    raise ImportError(
        "Using a jax objective requires an installation of "
        "the python package jax. Please install jax via "
        "`pip install jax jaxlib`."
    ) from None


def _base_objective_as_jax_array_tuple(func: Callable):
    def decorator(*args, **kwargs):
        # make sure return is a tuple of jax arrays
        results = func(*args, **kwargs)
        if isinstance(results, tuple):
            return tuple(jnp.array(r) for r in results)
        return jnp.array(results)

    return decorator


# jax compatible (jit-able) objective function using external callback, see
# https://jax.readthedocs.io/en/latest/notebooks/external_callbacks.html


@partial(custom_jvp, nondiff_argnums=(0,))
def _device_fun(base_objective: ObjectiveBase, x: jnp.array) -> jnp.array:
    """Jax compatible objective function execution using external callback.

    Parameters
    ----------
    obj:
        The wrapped jax objective.
    x:
        jax computed input array.

    Returns
    -------
    fval : jnp.array
        The function value as 0-dimensional jax array.
    """
    return jax.pure_callback(
        _base_objective_as_jax_array_tuple(
            partial(base_objective, sensi_orders=(0,))
        ),
        jax.ShapeDtypeStruct((), x.dtype),
        x,
        vmap_method="sequential",
    )


def _device_fun_value_and_grad(
    base_objective: ObjectiveBase, x: jnp.array
) -> tuple[jnp.array, jnp.array]:
    """Jax compatible objective gradient execution using external callback.

    This function will be called when computing the gradient of the
    `JaxObjective` using `jax.grad` or `jax.value_and_grad`. In the latter
    case, the function will return both the function value and the gradient,
    so no caching is necessary. For higher order derivatives, caching would
    be advantageous, but unclear how to implement this.

    Parameters
    ----------
    obj:
        The wrapped jax objective.
    x:
        jax computed input array.

    Returns
    -------
    fval : jnp.array
        The function value as 0-dimensional jax array.
    grad : jnp.array
        The gradient as jax array.
    """
    return jax.pure_callback(
        _base_objective_as_jax_array_tuple(
            partial(
                base_objective,
                sensi_orders=(
                    0,
                    1,
                ),
            )
        ),
        (
            jax.ShapeDtypeStruct((), x.dtype),
            jax.ShapeDtypeStruct(
                x.shape,  # bootstrap from cached value
                x.dtype,
            ),
        ),
        x,
        vmap_method="sequential",
    )


# define custom jvp for device_fun to enable autodiff, see
# https://jax.readthedocs.io/en/latest/notebooks/Custom_derivative_rules_for_Python_code.html


@_device_fun.defjvp
def _device_fun_jvp(
    obj: "JaxObjective", primals: jnp.array, tangents: jnp.array
):
    """JVP implementation for device_fun."""
    (x,) = primals
    (x_dot,) = tangents
    value, grad = _device_fun_value_and_grad(obj, x)
    return value, grad @ x_dot


class JaxObjective(ObjectiveBase):
    """Objective function that enables use of pypesto objectives in jax models.

    The generated function should generally be compatible with jax, but cannot
    compute higher order derivatives and is not vectorized (but still
    compatible with jax.vmap)

    Parameters
    ----------
    objective:
        pyPESTO objective to be wrapped.

    Note
    ----
    Currently only implements MODE_FUN and sensi_orders<=1. Support for
    MODE_RES should be straightforward to add.
    """

    def __init__(
        self,
        objective: ObjectiveBase,
    ):
        if not isinstance(objective, ObjectiveBase):
            raise TypeError("objective must be an ObjectiveBase instance")
        if not objective.check_mode(MODE_FUN):
            raise NotImplementedError(
                f"objective must support mode={MODE_FUN}"
            )
        self.base_objective = objective

        # would be cleaner to also have this as class method, but not supported
        # by signature inspection in bind call.
        self.jax_objective = partial(_device_fun, self.base_objective)

    def check_mode(self, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        return mode == MODE_FUN

    def check_sensi_orders(self, sensi_orders, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        if not self.check_mode(mode):
            return False
        else:
            return (
                self.base_objective.check_sensi_orders(sensi_orders, mode)
                and max(sensi_orders) <= 1
            )

    def __call__(
        self,
        x: jnp.ndarray,
        sensi_orders: tuple[int, ...] = (0,),
        mode: ModeType = MODE_FUN,
        return_dict: bool = False,
        **kwargs,
    ) -> jnp.ndarray | tuple | ResultDict:
        """
        See :class:`ObjectiveBase` for more documentation.

        Note that this function delegates pre- and post-processing as well as
        history handling to the inner objective.
        """

        if not self.check_mode(mode):
            raise ValueError(
                f"This Objective cannot be called with mode={mode}."
            )
        if not self.check_sensi_orders(sensi_orders, mode):
            raise ValueError(
                f"This Objective cannot be called with "
                f"sensi_orders= {sensi_orders} and mode={mode}."
            )

        # this computes all the results from the inner objective, rendering
        # them accessible as cached values for device_fun, etc.
        if kwargs.pop("return_dict", False):
            raise ValueError(
                "return_dict=True is not available for JaxObjective evaluation"
            )

        return self.jax_objective(x)

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: tuple[int, ...],
        mode: ModeType,
        return_dict: bool,
        **kwargs,
    ) -> ResultDict:
        """
        See :class:`ObjectiveBase` for more documentation.

        This function is not implemented for JaxObjective as it is not called
        in the override for __call__. However, it's marked as abstract so we
        need to implement it.
        """
        pass

    def __deepcopy__(self, memodict=None):
        other = JaxObjective(
            copy.deepcopy(self.base_objective),
        )
        return other

    @property
    def history(self):
        """Expose the history of the inner objective."""
        return self.base_objective.history

    @property
    def pre_post_processor(self):
        """Expose the pre_post_processor of inner objective."""
        return self.base_objective.pre_post_processor

    @pre_post_processor.setter
    def pre_post_processor(self, new_pre_post_processor):
        """Set the pre_post_processor of inner objective."""
        self.base_objective.pre_post_processor = new_pre_post_processor

    @property
    def x_names(self):
        """Expose the x_names of inner objective."""
        return self.base_objective.x_names
